########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


import cv2
import matplotlib
import matplotlib.colors
import numpy as np
import torch
import torchvision
import torchvision.transforms as transforms


def get_preprocessor(
    height=None,
    width=None,
    phase="train",
    train_random_rescale=(1.0, 1.4),
):
    assert phase in ["train", "test"]

    if phase == "train":
        transform_list = [
            RandomRescale(train_random_rescale),
            RandomCrop(crop_height=height, crop_width=width),
            RandomHSV((0.9, 1.1), (0.9, 1.1), (25, 25)),
            RandomFlip(),
            ToTensor(),
            Normalize(),
        ]

    else:
        if height is None and width is None:
            transform_list = []
        else:
            transform_list = [Rescale(height=height, width=width)]
        transform_list.extend(
            [
                ToTensor(),
                Normalize(),
            ]
        )
    transform = transforms.Compose(transform_list)
    return transform


class Rescale:
    def __init__(self, height, width):
        self.height = height
        self.width = width

    def __call__(self, sample):
        image = sample["image"]

        image = cv2.resize(
            image, (self.width, self.height), interpolation=cv2.INTER_LINEAR
        )

        sample["image"] = image

        if "label" in sample:
            label = sample["label"]
            label = cv2.resize(
                label, (self.width, self.height), interpolation=cv2.INTER_NEAREST
            )
            sample["label"] = label

        return sample


class RandomRescale:
    def __init__(self, scale):
        self.scale_low = min(scale)
        self.scale_high = max(scale)

    def __call__(self, sample):
        image, label = sample["image"], sample["label"]

        target_scale = np.random.uniform(self.scale_low, self.scale_high)
        # (H, W, C)
        target_height = int(round(target_scale * image.shape[0]))
        target_width = int(round(target_scale * image.shape[1]))

        image = cv2.resize(
            image, (target_width, target_height), interpolation=cv2.INTER_LINEAR
        )
        label = cv2.resize(
            label, (target_width, target_height), interpolation=cv2.INTER_NEAREST
        )
        sample["image"] = image
        sample["label"] = label

        return sample


class RandomCrop:
    def __init__(self, crop_height, crop_width):
        self.crop_height = crop_height
        self.crop_width = crop_width
        self.rescale = Rescale(self.crop_height, self.crop_width)

    def __call__(self, sample):
        image, label = sample["image"], sample["label"]
        h = image.shape[0]
        w = image.shape[1]
        if h <= self.crop_height or w <= self.crop_width:
            # simply rescale instead of random crop as image is not large enough
            sample = self.rescale(sample)
        else:
            i = np.random.randint(0, h - self.crop_height)
            j = np.random.randint(0, w - self.crop_width)
            image = image[i : i + self.crop_height, j : j + self.crop_width, :]
            label = label[i : i + self.crop_height, j : j + self.crop_width]
            sample["image"] = image
            sample["label"] = label
        return sample


class RandomHSV:
    def __init__(self, h_range, s_range, v_range):
        assert (
            isinstance(h_range, (list, tuple))
            and isinstance(s_range, (list, tuple))
            and isinstance(v_range, (list, tuple))
        )
        self.h_range = h_range
        self.s_range = s_range
        self.v_range = v_range

    def __call__(self, sample):
        img = sample["image"]
        img_hsv = matplotlib.colors.rgb_to_hsv(img)
        img_h = img_hsv[:, :, 0]
        img_s = img_hsv[:, :, 1]
        img_v = img_hsv[:, :, 2]

        h_random = np.random.uniform(min(self.h_range), max(self.h_range))
        s_random = np.random.uniform(min(self.s_range), max(self.s_range))
        v_random = np.random.uniform(-min(self.v_range), max(self.v_range))
        img_h = np.clip(img_h * h_random, 0, 1)
        img_s = np.clip(img_s * s_random, 0, 1)
        img_v = np.clip(img_v + v_random, 0, 255)
        img_hsv = np.stack([img_h, img_s, img_v], axis=2)
        img_new = matplotlib.colors.hsv_to_rgb(img_hsv)

        sample["image"] = img_new

        return sample


class RandomFlip:
    def __call__(self, sample):
        image, label = sample["image"], sample["label"]
        if np.random.rand() > 0.5:
            image = np.fliplr(image).copy()
            label = np.fliplr(label).copy()

        sample["image"] = image
        sample["label"] = label

        return sample


class Normalize:
    def __call__(self, sample):
        image = sample["image"]
        image = image / 255
        image = torchvision.transforms.Normalize(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )(image)

        sample["image"] = image

        return sample


class ToTensor:
    def __call__(self, sample):
        image = sample["image"]
        image = image.transpose((2, 0, 1))

        sample["image"] = torch.from_numpy(image).float()

        if "label" in sample:
            label = sample["label"]
            sample["label"] = torch.from_numpy(label).float()

        return sample


class MultiScaleLabel:
    def __init__(self, downsampling_rates=None):
        if downsampling_rates is None:
            self.downsampling_rates = [8, 16, 32]
        else:
            self.downsampling_rates = downsampling_rates

    def __call__(self, sample):
        label = sample["label"]

        h, w = label.shape

        sample["label_down"] = dict()

        # Nearest neighbor interpolation
        for rate in self.downsampling_rates:
            label_down = cv2.resize(
                label.numpy(), (w // rate, h // rate), interpolation=cv2.INTER_NEAREST
            )
            sample["label_down"][rate] = torch.from_numpy(label_down)

        return sample



# from .datasets import CustomDOTADataset, CustomHRSCDataset
# from .models.backbones import ReResNet, VisionTransformer, HiViT
# from .models.roi_heads import (RotatedBBoxHead, RotatedConvFCBBoxHead, RotatedShared2FCBBoxHead, 
#                                 RotatedStandardRoIHead, RotatedSingleRoIExtractor,
#                                 OrientedStandardRoIHead, RoITransRoIHead, GVRatioRoIHead,
#                                 RotatedStandardRoIHeadimTED,
#                                 OrientedStandardRoIHeadimTED)
# from .models.roi_heads.bbox_heads import (RotatedBBoxHead, RotatedConvFCBBoxHead, RotatedShared2FCBBoxHead,
#                                             GVBBoxHead, RotatedKFIoUShared2FCBBoxHead,
#                                             RotatedMAEBBoxHead, RotatedMAEBBoxHeadSTDC)
# from .models.necks import ReFPN, SimpleFPN
# from .models.detectors import (RotatedRetinaNet, RotatedFasterRCNN, OrientedRCNN, RoITransformer,
#                                 GlidingVertex, ReDet, R3Det, S2ANet, RotatedRepPoints,
#                                 RotatedBaseDetector, RotatedTwoStageDetector,
#                                 RotatedSingleStageDetector, RotatedFCOS,
#                                 RotatedimTED)

from .datasets import *
from .models.backbones import *
from .models.roi_heads import *
from .models.roi_heads.bbox_heads import *
from .models.necks import *
from .models.detectors import *


#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule


class GroupedConvFuse(nn.Module):
    """Basic multi-layer side fusion used in CaseNet

    https://github.com/Lavender105/DFF/blob/master/exps/models/casenet.py

    Changes:
    - bias=False: no bias in the last layer
    - flexible: number of sides could channge (CASENet, DDS, etc...)
    """

    def __init__(
        self,
        num_classes,
        num_sides,
        conv_cfg=None,
        bias=True,
    ):
        super().__init__()

        self.num_sides = num_sides

        # fuse (grouped convolution)
        self.fuse = ConvModule(
            in_channels=num_classes * num_sides,
            out_channels=num_classes,
            kernel_size=1,
            groups=num_classes,
            conv_cfg=conv_cfg,
            norm_cfg=None,
            bias=bias,  # originally True
            act_cfg=None,
        )

    def forward(self, sides):
        assert isinstance(sides, list)
        assert len(sides) == self.num_sides, f"number of sides: {len(sides)}"

        # fixed inplace array
        side5 = sides[-1]
        sides = sides[:-1]

        slice5 = side5[:, 0:1, :, :]
        fuse = torch.cat((slice5, *sides), 1)
        for i in range(side5.size(1) - 1):
            slice5 = side5[:, i + 1 : i + 2, :, :]
            fuse = torch.cat((fuse, slice5, *sides), 1)

        fuse = self.fuse(fuse)

        return fuse


class GeneralizedLocationAdaptiveLearner(nn.Module):
    def __init__(
        self,
        num_sides,
        in_channels,
        out_channels,
        conv_cfg,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
    ):
        super().__init__()

        self.num_sides = num_sides

        self.conv_layers = nn.Sequential(
            nn.ReLU(),  # NOTE: added activation
            ConvModule(
                in_channels,
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
            ),
            ConvModule(
                out_channels,
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
            ),
            ConvModule(
                out_channels,
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
            nn.Sigmoid(),
        )

    def forward(self, sides):
        assert isinstance(sides, list)
        assert len(sides) == self.num_sides + 1, f"number of sides: {len(sides)}"

        num_backbones = len(sides) - 1

        side_w = sides[-1]
        side_l = sides[-2]
        sides = sides[:-2]

        num_classes = side_l.size(1)

        slice_l = side_l[:, 0:1, :, :]
        fuse = torch.cat((slice_l, *sides), 1)
        for i in range(side_l.size(1) - 1):
            slice_l = side_l[:, i + 1 : i + 2, :, :]
            fuse = torch.cat((fuse, slice_l, *sides), 1)

        # (N, 19*4, H, W)
        w = self.conv_layers(side_w)
        # Reshape to (N, 19, 4, H, W)
        w = w.view(w.size(0), num_classes, num_backbones, w.size(2), w.size(3))
        fuse = fuse.view(fuse.size(0), num_classes, -1, fuse.size(2), fuse.size(3))

        # fuse = torch.mul(fuse, w)
        # fuse = torch.sum(fuse, 2)
        # fuse = fuse * torch.sigmoid(w)
        fuse = fuse * w
        fuse = torch.sum(fuse, 2)

        return fuse



import time
import torch
from abc import abstractmethod
import os.path as osp
from pathlib import Path 
import os 

from visionsuite.engines.utils.metrics.metric_logger import MetricLogger
from visionsuite.engines.utils.metrics.smoothed_value import SmoothedValue
from visionsuite.engines.segmentation.utils.registry import TESTERS
from visionsuite.engines.utils.system.gpu_logger import GPULogger
from visionsuite.engines.utils.bases import BaseOOPModule
from visionsuite.engines.utils.callbacks import Callbacks
from visionsuite.engines.utils.helpers import increment_path
from visionsuite.engines.segmentation.utils.vis.vis_test import vis_by_batch
from visionsuite.engines.utils.functionals import denormalize
from .callbacks import callbacks


@TESTERS.register()
class BaseTester(BaseOOPModule, Callbacks):
    
    required_attributes = ['model', 'loss', 'dataloader']
    
    def __init__(self, name='BaseTester'):
        BaseOOPModule.__init__(self, name=name)
        Callbacks.__init__(self)
        
        self.add_callbacks(callbacks)
        
    def build(self, model, loss, dataloader, scaler=None, archive=None, **args):
        super().build(**args)
        self.run_callbacks('on_tester_build_start')
        if ('output_dir' in self.args and self.args['output_dir'] is None) or 'output_dir' not in self.args:
            output_dir = osp.join(osp.splitext(self.args['seed_model'])[0], '../../../test')
            
            if not osp.exists(output_dir):
                os.makedirs(output_dir)
            output_dir = str(increment_path(Path(output_dir) / "exp", exist_ok=False, mkdir=True))
            self.args['output_dir'] = output_dir
            
        assert osp.exists(self.args['output_dir']), ValueError(f"There is no such output-dir: {self.args['output_dir']}")
        
        self.epoch = None
        
        self.model = model
        self.loss = loss
        self.dataloader = dataloader
        self.scaler = scaler
        self.archive = archive
        
        self.metric_logger = MetricLogger(delimiter="  ")
        self.log_info(f"SET MetricLogger", self.build.__name__, __class__.__name__)
        
        self.gpu_logger = GPULogger(self.args['device_ids'])
        self.log_info(f"SET GPULogger with devices({self.args['device_ids']})", self.build.__name__, __class__.__name__)
        
        self.run_callbacks('on_tester_build_end')

    @abstractmethod
    def run(self, epoch):
        self.run_callbacks('on_tester_epoch_start')
        start_time_epoch = time.time()
        self.model.model.eval()
        header = f"Epoch: [{epoch}]"
        self.log_info(f"START test epoch: {epoch}", self.run.__name__, __class__.__name__)
        
        for batch_idx, batch in enumerate(self.metric_logger.log_every(self.dataloader, self.args['print_freq'], header)):
            self.run_callbacks('on_tester_batch_start')
            start_time = time.time()
            image, target, fname = batch[0].to(self.args['device']), batch[1].to(self.args['device']), batch[2]
            self.log_debug(f"- image: {image.shape} with device({self.args['device']})", self.run.__name__, __class__.__name__)
            self.log_debug(f"- target: {target.shape} with device({self.args['device']})", self.run.__name__, __class__.__name__)
            with torch.cuda.amp.autocast(enabled=self.scaler is not None):
                output = self.model.model(image) # ';out': (bs num_classes(including bg) h w)
                loss = self.loss(output, target)

            vis_by_batch(output['out'], target, image, fname, epoch, batch_idx, output_dir=self.args['output_dir'],
                         denormalize=denormalize)
                
            self._update_logger(output, target, loss, start_time, batch_size=image.shape[0])
            
            self.run_callbacks('on_tester_batch_end')

        self.run_callbacks('on_tester_epoch_end', 
                           epoch=epoch, start_time_epoch=start_time_epoch)
    
                
    def _update_logger(self, output, target, loss, start_time, batch_size):
        if self.metric_logger is not None:
            self.metric_logger.meters["img/s"].update(batch_size / (time.time() - start_time))
            self.metric_logger.meters['gpu'].update((torch.cuda.memory_allocated() + torch.cuda.memory_reserved()) / 1024**2)
        
            self.log_debug(f"- Updated metric_logger", self._update_logger.__name__, __class__.__name__)
        if self.gpu_logger is not None:
            self.gpu_logger.update()
        
            self.log_debug(f"- Updated gpu_logger", self._update_logger.__name__, __class__.__name__)


import argparse
import logging
import os
import numpy as np
import torch
import torch.nn.functional as F
import torchsummary as summary

from PIL import Image
from torchvision import transforms
from unet import UNet
from unet import UNet2Plus
from unet import UNet3Plus, UNet3Plus_DeepSup, UNet3Plus_DeepSup_CGM
from utils.data_vis import plot_img_and_mask
from utils.dataset import BasicDataset


def predict_img(unet_type, net, full_img, device, scale_factor=1, out_threshold=0.5):
    net.eval()
    img = torch.from_numpy(BasicDataset.preprocess(unet_type, full_img, scale_factor))
    img = img.unsqueeze(0)
    img = img.to(device=device, dtype=torch.float32)

    with torch.no_grad():
        output = net(img)
        if net.n_classes > 1:
            probs = F.softmax(output, dim=1)
        else:
            probs = torch.sigmoid(output)

        probs = probs.squeeze(0)
        tf = transforms.Compose([transforms.ToPILImage(), transforms.Resize(full_img.size[1]),
                                 transforms.ToTensor()])
        probs = tf(probs.cpu())
        full_mask = probs.squeeze().cpu().numpy()
    return full_mask > out_threshold


def get_args():
    parser = argparse.ArgumentParser(description='Predict masks from input images',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--gpu_id', '-g', metavar='G', type=int, default=0, help='Number of gpu')
    parser.add_argument('--unet_type', '-u', metavar='U', default='v3', help='UNet type is v1/v2/v3 (unet unet++ unet3+)')
    parser.add_argument('--model', '-m', default='MODEL.pth', metavar='FILE', help='Specify the file in which the model is stored')
    
    parser.add_argument('--input', '-i', metavar='INPUT', nargs='+', help='filenames of input images', required=True)
    parser.add_argument('--output', '-o', metavar='OUTPUT', nargs='+', help='Filenames of ouput images')
    
    parser.add_argument('--viz', '-v', action='store_true', help='Visualize the images as they are processed', default=False)
    parser.add_argument('--no-save', '-n', action='store_true', help='Do not save the output masks', default=False)
    
    parser.add_argument('--mask-threshold', '-t', type=float, help='Minimum probability value to consider a mask pixel white', default=0.5)
    parser.add_argument('--scale', '-s', type=float, help='Scale factor for the input images', default=0.5)
    return parser.parse_args()


def get_output_filenames(args):
    in_files = args.input
    out_files = []

    if not args.output:
        for f in in_files:
            pathsplit = os.path.splitext(f)
            out_files.append('{}_OUT{}'.format(pathsplit[0], pathsplit[1]))
    elif len(in_files) != len(args.output):
        logging.error('Input files and output files are not of the same length')
        raise SystemExit()
    else:
        out_files = args.output
    return out_files


def mask_to_image(mask):
    return Image.fromarray((mask * 255).astype(np.uint8))


if __name__ == '__main__':
    args = get_args()
    gpu_id = args.gpu_id 
    unet_type = args.unet_type    
    in_files = args.input
    out_files = get_output_filenames(args)

    os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu_id)
    if unet_type == 'v2':
        net = UNet2Plus(n_channels=3, n_classes=1)
    elif unet_type == 'v3':
        net = UNet3Plus(n_channels=3, n_classes=1)
        #net = UNet3Plus_DeepSup(n_channels=3, n_classes=1)
        #net = UNet3Plus_DeepSup_CGM(n_channels=3, n_classes=1)
    else:
        net = UNet(n_channels=3, n_classes=1)

    logging.info('Loading model {}'.format(args.model))
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    
    logging.info(f'Using device {device}')
    net.to(device=device)
    net.load_state_dict(torch.load(args.model, map_location=device))
    logging.info('Model loaded !')

    for i, fn in enumerate(in_files):
        logging.info('\nPredicting image {} ...'.format(fn))
        img = Image.open(fn)
        mask = predict_img(unet_type=unet_type, net=net, full_img=img, scale_factor=args.scale,
                           out_threshold=args.mask_threshold, device=device)

        if not args.no_save:
            out_fn = out_files[i]
            result = mask_to_image(mask)
            result.save(out_files[i])
            logging.info('Mask saved to {}'.format(out_files[i]))

        if args.viz:
            logging.info('Visualizing results for image {}, close to continue ...'.format(fn))
            plot_img_and_mask(img, mask)



import argparse
from datetime import datetime
import json
import os
import sys
import time
import numpy as np
from tqdm import tqdm

import torch
from torch.utils.tensorboard import SummaryWriter
import torch.optim
from torch.optim.lr_scheduler import OneCycleLR
from src.args import ArgumentParser
from src.build_model import build_model
from src import utils
from src.prepare_data import prepare_data
from src.utils import save_ckpt_every_epoch
from src.utils import load_ckpt
from src.utils import print_log
import imgviz
import cv2
import os.path as osp

from torchmetrics import JaccardIndex as IoU


def parse_args():
    parser = ArgumentParser(
        description="Open-World Semantic Segmentation (Training)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.set_common_args()
    args = parser.parse_args()
    # The provided learning rate refers to the default batch size of 8.
    # When using different batch sizes we need to adjust the learning rate
    # accordingly:
    if args.batch_size != 8:
        args.lr = args.lr * args.batch_size / 8
        print(
            f"Notice: adapting learning rate to {args.lr} because provided "
            f"batch size differs from default batch size of 8."
        )

    return args


def train_main():
    args = parse_args()

    # directory for storing weights and other training related files
    training_starttime = datetime.now().strftime("%d_%m_%Y-%H_%M_%S-%f")
    ckpt_dir = os.path.join(
        args.results_dir, args.dataset, f"{args.id}", f"{training_starttime}"
    )
    vis_dir = os.path.join(
        args.results_dir, args.dataset, f"{args.id}", f"{training_starttime}", 'vis'
    )
    os.makedirs(ckpt_dir, exist_ok=True)
    os.makedirs(vis_dir, exist_ok=True)
    os.makedirs(os.path.join(ckpt_dir, "confusion_matrices"), exist_ok=True)

    with open(os.path.join(ckpt_dir, "args.json"), "w") as f:
        json.dump(vars(args), f, sort_keys=True, indent=4)

    with open(os.path.join(ckpt_dir, "argsv.txt"), "w") as f:
        f.write(" ".join(sys.argv))
        f.write("\n")

    # data preparation ---------------------------------------------------------
    data_loaders = prepare_data(args, ckpt_dir)

    train_loader, valid_loader, _ = data_loaders

    n_classes_without_void = train_loader.dataset.n_classes_without_void
    if args.class_weighting != "None":
        class_weighting = train_loader.dataset.compute_class_weights(
            weight_mode=args.class_weighting, c=args.c_for_logarithmic_weighting
        )
    else:
        class_weighting = np.ones(n_classes_without_void)
    # model building -----------------------------------------------------------
    model, device = build_model(args, n_classes=n_classes_without_void)
    if args.freeze > 0:
        print("Freeze everything but the output layer(s).")
        for name, param in model.named_parameters():
            if "out" not in name:
                param.requires_grad = False

    # loss, optimizer, learning rate scheduler, csvlogger  ----------

    # loss functions
    loss_function_train = utils.CrossEntropyLoss2d(
        weight=class_weighting, device=device
    )
    loss_objectosphere = utils.ObjectosphereLoss()
    loss_mav = utils.OWLoss(n_classes=n_classes_without_void)
    loss_contrastive = utils.ContrastiveLoss(n_classes=n_classes_without_void)

    pixel_sum_valid_data = valid_loader.dataset.compute_class_weights(
        weight_mode="linear"
    )
    pixel_sum_valid_data_weighted = np.sum(pixel_sum_valid_data * class_weighting)
    loss_function_valid = utils.CrossEntropyLoss2dForValidData(
        weight=class_weighting,
        weighted_pixel_sum=pixel_sum_valid_data_weighted,
        device=device,
    )

    train_loss = [loss_function_train, loss_objectosphere, loss_mav, loss_contrastive]
    val_loss = [loss_function_valid, loss_objectosphere, loss_mav, loss_contrastive]
    if not args.obj:
        train_loss[1] = None
        val_loss[1] = None
    if not args.mav:
        train_loss[2] = None
        val_loss[2] = None
    if not args.closs:
        train_loss[3] = None
        val_loss[3] = None

    optimizer = get_optimizer(args, model)

    # in this script lr_scheduler.step() is only called once per epoch
    lr_scheduler = OneCycleLR(
        optimizer,
        max_lr=[i["lr"] for i in optimizer.param_groups],
        total_steps=args.epochs,
        div_factor=25,
        pct_start=0.1,
        anneal_strategy="cos",
        final_div_factor=1e4,
    )

    # load checkpoint if parameter last_ckpt is provided
    if args.last_ckpt:
        ckpt_path = args.last_ckpt
        epoch_last_ckpt, best_miou, best_miou_epoch, mav_dict, std_dict = load_ckpt(
            model, optimizer, ckpt_path, device
        )
        start_epoch = epoch_last_ckpt + 1
        print(">>>> LOADED ckpt: ", args.last_ckpt)
    else:
        start_epoch = 0
        best_miou = 0
        best_miou_epoch = 0

    if args.load_weights:
        model.load_state_dict(torch.load(args.load_weights))
        print(">>>> LOADED weights: ", args.load_weights)

    writer = SummaryWriter("runs/" + ckpt_dir.split(args.dataset)[-1])
    color_map = imgviz.label_colormap(50)
    # start training -----------------------------------------------------------
    for epoch in range(int(start_epoch), args.epochs):
        # unfreeze
        if args.freeze == epoch and args.finetune is None:
            for param in model.parameters():
                param.requires_grad = True

        mean, var = train_one_epoch(
            model=model,
            train_loader=train_loader,
            device=device,
            optimizer=optimizer,
            train_loss=train_loss,
            epoch=epoch,
            lr_scheduler=lr_scheduler,
            debug_mode=args.debug,
            writer=writer,
        )

        miou = validate(
            model=model,
            valid_loader=valid_loader,
            device=device,
            val_loss=val_loss,
            epoch=epoch,
            vis_dir=vis_dir,
            color_map=color_map,
            debug_mode=args.debug,
            writer=writer,
            classes=args.num_classes,
        )

        writer.flush()

        # save weights
        if not args.overfit:
            # save / overwrite latest weights (useful for resuming training)
            save_ckpt_every_epoch(
                ckpt_dir, model, optimizer, epoch, best_miou, best_miou_epoch, mean, var
            )
            if (epoch + 1) % 20 == 0:
                torch.save(
                    model.state_dict(),
                    os.path.join(ckpt_dir, "epoch_" + str(epoch) + ".pth"),
                )
            if miou > best_miou:
                best_miou = miou
                best_miou_epoch = epoch
                torch.save(
                    model.state_dict(),
                    os.path.join(ckpt_dir, "best_miou.pth"),
                )

    # save mavs to a pickle
    with open("mavs.pickle", "wb") as h1:
        pickle.dump(mean, h1, protocol=pickle.HIGHEST_PROTOCOL)
    with open("vars.pickle", "wb") as h2:
        pickle.dump(var, h2, protocol=pickle.HIGHEST_PROTOCOL)

    print("Training completed ")


def train_one_epoch(
    model,
    train_loader,
    device,
    optimizer,
    train_loss,
    epoch,
    lr_scheduler,
    writer,
    debug_mode=False,
):
    lr_scheduler.step(epoch)
    samples_of_epoch = 0

    # set model to train mode
    model.train()

    loss_function_train, loss_obj, loss_mav, loss_contrastive = train_loss

    # summed loss of all resolutions
    total_loss_list = []
    total_sem_loss = []
    total_obj_loss = []
    total_ows_loss = []
    total_con_loss = []

    mavs = None
    if epoch and loss_contrastive is not None:
        mavs = loss_mav.read()
    for i, sample in enumerate(train_loader):
        start_time_for_one_step = time.time()

        # load the data and send them to gpu
        image = sample["image"].to(device)
        batch_size = image.data.shape[0]

        label_ss = sample["label"].clone().cuda()
        label_ss[label_ss == 255] = 0
        target_scales = label_ss

        for param in model.parameters():
            param.grad = None

        # forward pass
        pred_scales, ow_res = model(image)
        cw_target = target_scales.clone()
        # cw_target[cw_target > 16] = 255
        losses = loss_function_train(pred_scales, cw_target)
        loss_segmentation = sum(losses)
        loss_objectosphere = torch.tensor(0.0)
        loss_ows = torch.tensor(0.0)
        loss_con = torch.tensor(0.0)
        total_loss = 0.9 * loss_segmentation
        label = sample["label"].long().cuda() - 1
        label[label < 0] = 255

        if loss_obj is not None:
            label_ow = label.clone().cuda().to(torch.uint8)
            loss_objectosphere = loss_obj(ow_res, label_ow)
            total_loss += 0.5 * loss_objectosphere
        if loss_mav is not None:
            loss_ows = loss_mav(pred_scales, label, is_train=True)
            total_loss += 0.1 * loss_ows
        if loss_contrastive is not None:
            loss_con = loss_contrastive(mavs, ow_res, label, epoch)
            total_loss += 0.5 * loss_con

        total_loss.backward()
        optimizer.step()

        # append loss values to the lists. Later we can calculate the
        # mean training loss of this epoch
        total_loss = total_loss.cpu().detach().numpy()
        loss_segmentation = loss_segmentation.cpu().detach().numpy()
        loss_objectosphere = loss_objectosphere.cpu().detach().numpy()
        loss_ows = loss_ows.cpu().detach().numpy()
        loss_con = loss_con.cpu().detach().numpy()

        total_loss_list.append(total_loss)
        total_sem_loss.append(loss_segmentation)
        total_obj_loss.append(loss_objectosphere)
        total_ows_loss.append(loss_ows)
        total_con_loss.append(loss_con)

        if np.isnan(total_loss):
            import ipdb;ipdb.set_trace()  # fmt: skip
            raise ValueError("Loss is None")

        # print log
        samples_of_epoch += batch_size
        time_inter = time.time() - start_time_for_one_step

        learning_rates = lr_scheduler.get_lr()

        print_log(
            epoch,
            samples_of_epoch,
            batch_size,
            len(train_loader.dataset),
            total_loss,
            time_inter,
            learning_rates,
        )

        if debug_mode:
            # only one batch while debugging
            break

    # fill the logs for csv log file and web logger
    writer.add_scalar("Loss/train", np.mean(total_loss_list), epoch)
    writer.add_scalar("Loss/semantic", np.mean(total_sem_loss), epoch)
    writer.add_scalar("Loss/objectosphere", np.mean(total_obj_loss), epoch)
    writer.add_scalar("Loss/ows", np.mean(total_ows_loss), epoch)
    writer.add_scalar("Loss/contrastive", np.mean(total_con_loss), epoch)

    if loss_mav is not None:
        mean, var = loss_mav.update()
        return mean, var
    else:
        return {}, {}


def validate(
    model,
    valid_loader,
    device,
    val_loss,
    epoch,
    writer,
    vis_dir,
    color_map,
    loss_function_valid_unweighted=None,
    debug_mode=False,
    classes=19,
):
    # set model to eval mode
    model.eval()

    # we want to store miou and ious for each camera
    miou = dict()
    ious = dict()

    loss_function_valid, loss_obj, loss_mav, loss_contrastive = val_loss

    # reset loss (of last validation) to zero
    loss_function_valid.reset_loss()

    if loss_function_valid_unweighted is not None:
        loss_function_valid_unweighted.reset_loss()

    compute_iou = IoU(
        task="multiclass", num_classes=classes, average="none", ignore_index=255
    ).to(device)

    mavs = None
    if loss_contrastive is not None:
        mavs = loss_mav.read()

    total_loss_obj = []
    total_loss_mav = []
    total_loss_con = []
    # validate each camera after another as all images of one camera have
    # the same resolution and can be resized together to the ground truth
    # segmentation size.

    cnt = 1
    for i, sample in enumerate(tqdm(valid_loader, desc="Valid step")):
        # copy the data to gpu
        image = sample["image"].to(device)

        if not device.type == "cpu":
            torch.cuda.synchronize()

        # forward pass
        with torch.no_grad():
            prediction_ss, prediction_ow = model(image)
            target = sample["label"].long().cuda() - 1
            target[target == -1] = 255


            if epoch % 5 == 0:
                if not osp.exists(osp.join(vis_dir, str(epoch))):
                    os.mkdir(osp.join(vis_dir, str(epoch)))
                for (_image, _target, pred_ss) in zip(image, target, prediction_ss):
                    vis_image = np.zeros((512, 1024*3, 3))
                    
                    vis_pred_ss = pred_ss.detach().cpu().numpy()
                    vis_pred_ss = np.transpose(vis_pred_ss, (1, 2, 0))#.astype(np.uint8)
                    vis_pred_ss = np.argmax(vis_pred_ss, axis=-1)
                    vis_pred_ss = color_map[vis_pred_ss].astype(np.uint8)
                    
                    vis_image[:, :1024, :] = np.transpose(_image.detach().cpu().numpy(), (1, 2, 0))
                    vis_image[:, 1024:1024*2, :] = color_map[_target.detach().cpu().numpy()].astype(np.uint8)
                    vis_image[:, 1024*2:, :] = vis_pred_ss
                    cv2.imwrite(osp.join(vis_dir, str(epoch), f"{epoch}_{cnt}.png"), vis_image)
                    cnt += 1

            if not device.type == "cpu":
                torch.cuda.synchronize()

            compute_iou.update(prediction_ss, target.cuda())

            # compute valid loss
            loss_function_valid.add_loss_of_batch(
                prediction_ss, sample["label"].to(device)
            )

            loss_objectosphere = torch.tensor(0)
            loss_ows = torch.tensor(0)
            loss_con = torch.tensor(0)
            if loss_obj is not None:
                target_obj = sample["label"]
                target_obj[target_obj == 16] = 255
                target_obj[target_obj == 17] = 255
                target_obj[target_obj == 18] = 255
                loss_objectosphere = loss_obj(prediction_ow, sample["label"])
            total_loss_obj.append(loss_objectosphere.cpu().detach().numpy())
            if loss_mav is not None:
                loss_ows = loss_mav(prediction_ss, target.cuda(), is_train=False)
            total_loss_mav.append(loss_ows.cpu().detach().numpy())
            if loss_contrastive is not None:
                loss_con = loss_contrastive(mavs, prediction_ow, target, epoch)
            total_loss_con.append(loss_con.cpu().detach().numpy())

            if debug_mode:
                # only one batch while debugging
                break

    ious = compute_iou.compute().detach().cpu()
    miou = ious.mean()

    total_loss = (
        loss_function_valid.compute_whole_loss()
        + np.mean(total_loss_obj)
        + np.mean(total_loss_mav)
        + np.mean(total_loss_con)
    )
    writer.add_scalar("Loss/val", total_loss, epoch)
    writer.add_scalar("Metrics/miou", miou, epoch)
    for i, iou in enumerate(ious):
        writer.add_scalar(
            "Class_metrics/iou_{}".format(i),
            torch.mean(iou),
            epoch,
        )

    return miou


def test_ow(
    model,
    test_loader,
    device,
    val_loss,
    epoch,
    writer,
    classes=19,
    mean=None,
    var=None,
):
    delta = 0.6

    # set model to eval mode
    model.eval()

    compute_iou = IoU(
        task="multiclass", num_classes=2, average="none", ignore_index=255
    ).to(device)

    _, loss_obj, loss_mav, _ = val_loss

    with open("mavs.pickle", "rb") as h1:
        mavs = pickle.load(h1)
    with open("vars.pickle", "rb") as h2:
        vars = pickle.load(h2)

    mavs = torch.vstack(tuple(mavs.values())).cpu()  # 19x19
    new_mavs = None

    for i, sample in enumerate(tqdm(test_loader, desc="Test step")):
        # copy the data to gpu
        image = sample["image"].to(device)
        label = sample["label"].to(device)

        if not device.type == "cpu":
            torch.cuda.synchronize()

        # forward pass
        with torch.no_grad():
            prediction, ow_pred = model(image)

            ows_target = label.long() - 1
            ows_target[ows_target < classes] = 0
            ows_binary_gt = ows_target.bool().long()

            s_cont = contrastive_inference(ow_pred)
            s_sem, similarity = semantic_inference(prediction, mavs, vars)
            s_unk = (s_cont + s_sem) / 2

            ows_binary_pred = (s_unk - delta).relu().bool().int()

            compute_iou.update(ows_binary_pred, ows_binary_gt)

            prediction = prediction.permute(1, 0, 2, 3)
            unk_pixels = prediction[:, :, ows_binary_pred == 0]

            tmp = torch.ones(unk_pixels.shape)
            if new_mavs is not None:
                for i in range(new_mavs.shape[0]):
                    mav = new_mavs[:, i].unsqueeze(1)
                    dist = torch.norm(unk_pixels - mav, dim=0)
                    dist = (dist < 0.5).int()
                    tmp[:, dist == 1] = 0
                    upd = torch.mean(unk_pixels[dist == 1], dim=0)
                    new_mavs[i, :] = (new_mavs[i, :] + upd) / 2
            preds = unk_pixels * tmp
            preds = torch.unique(preds, dim=1)
            if tmp.sum():
                preds = preds[:, 1:]

            clusters = ac(
                n_clusters=None, affinity="euclidean", distance_threshold=0.5
            ).fit(preds.cpu().numpy().T)
            groups = clusters.labels_

            nc = groups.max()
            for c in nc:
                new = preds[:, groups == c]
                new = torch.mean(torch.tensor(new), dim=1)
                if new_mavs is None:
                    new_mavs = new
                else:
                    new_mavs = torch.vstack((new_mavs, new))

    ious = compute_iou.compute().detach().cpu()
    writer.add_scalar("Metrics/OWS/known", ious[0], epoch)
    writer.add_scalar("Metrics/OWS/unknown", ious[1], epoch)


def contrastive_inference(predictions, radius=1.0):
    scores = F.relu(1 - torch.norm(predictions, dim=1) / radius)
    return scores


def semantic_inference(predictions, mavs, var):
    stds = torch.vstack(tuple(var.values())).cpu()  # 19x19
    d_pred = (
        predictions[:, None, ...] - mavs[None, :, :, None, None]
    )  # [8,1,19,h,w] - [1,19,19,1,1]
    d_pred_ = d_pred / (stds[None, :, :, None, None] + 1e-8)
    scores = torch.exp(-torch.einsum("bcfhw,bcfhw->bchw", d_pred_, d_pred) / 2)
    best = scores.max(dim=1)
    return 1 - best[0], best[1]


def get_optimizer(args, model):
    # set different learning rates fo different parts of the model
    # when using default parameters the whole model is trained with the same
    # learning rate
    if args.optimizer == "SGD":
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
            momentum=args.momentum,
            nesterov=True,
        )
    elif args.optimizer == "Adam":
        optimizer = torch.optim.Adam(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
            betas=(0.9, 0.999),
        )
    else:
        raise NotImplementedError(
            "Currently only SGD and Adam as optimizers are "
            "supported. Got {}".format(args.optimizer)
        )

    print("Using {} as optimizer".format(args.optimizer))
    print(
        "\n\n=========================================================================\n\n"
    )
    return optimizer


if __name__ == "__main__":
    train_main()



import torch
import torch.nn.functional as F
import matplotlib.pyplot as plt

from torch import nn
from torch import Tensor
from PIL import Image
from torchvision.transforms import Compose, Resize, ToTensor
from einops import rearrange, reduce, repeat
from einops.layers.torch import Rearrange, Reduce
from torchsummary import summary

img_size = 224
x = torch.randn(8, 3, img_size, img_size) # bchw
patch_size = 16
in_channels = 3
batch_size = 8

# ----------------------
patches = rearrange(x, 'b c (h s1) (w s2) -> b (h w) (s1 s2 c)', s1=patch_size, s2=patch_size)
print(patches.shape)

# ----------------------
emb_size = patch_size*patch_size*in_channels

projection = nn.Sequential(
    nn.Conv2d(in_channels, emb_size, kernel_size=patch_size, stride=patch_size), # behw
    Rearrange('b e h w -> b (h w) e') # b(hw)e
)

print(projection(x).shape)

# ----------------------

projected_x = projection(x)
print("projected x: ", projected_x.shape)

cls_token = nn.Parameter(torch.randn(1, 1, emb_size))
positions = nn.Parameter(torch.randn((img_size//patch_size)**2 + 1, emb_size))
print(f'cls token: {cls_token.shape}')
print(f'positions: {positions.shape}')

cls_tokens = repeat(cls_token, '() n e -> b n e', b=batch_size)
print(f'cls tokens: {cls_tokens.shape}')

cat_x = torch.cat([cls_tokens, projected_x], dim=1)
cat_x += positions 
print(f'cat_x: ', cat_x.shape)

# ----------------------

# ----------------------

# ----------------------

# ----------------------

# ----------------------

# ----------------------

# ----------------------

# ----------------------


# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta

import torch
from mmcv.runner import BaseModule
from mmdet.core import bbox2roi

from mmrotate.core import build_assigner, build_sampler, obb2xyxy, rbbox2result
from mmrotate.models.builder import (ROTATED_HEADS, build_head, build_roi_extractor,
                       build_shared_head)


import torch.nn as nn
from mmrotate.models.roi_heads.oriented_standard_roi_head import OrientedStandardRoIHead


@ROTATED_HEADS.register_module()
class OrientedStandardRoIHeadimTED(OrientedStandardRoIHead):
    """RoIHead with multi-scale feature modulator on the input of bbox head."""



    def init_bbox_head(self, bbox_roi_extractor, bbox_head):
        """Initialize ``bbox_head``"""
        if isinstance(bbox_roi_extractor, list):
            self.ms_bbox_roi_extractor = build_roi_extractor(bbox_roi_extractor[0])
            self.ss_bbox_roi_extractor = build_roi_extractor(bbox_roi_extractor[1])
            self.mfm_factor = nn.Parameter(torch.zeros(bbox_head['in_channels'], requires_grad=True), requires_grad=True)
            self.mfm_fc = nn.Conv2d(in_channels=256,out_channels=bbox_head['in_channels'],kernel_size=1)
            self.with_mfm = True
        else:
            self.bbox_roi_extractor = build_roi_extractor(bbox_roi_extractor)
            self.with_mfm = False
        self.bbox_head = build_head(bbox_head)



    def init_weights(self, pretrained=None):
        """Initialize the weights in head.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """
        if self.with_shared_head:
            # self.shared_head.init_weights(pretrained=pretrained)
            self.shared_head.init_weights()
        if self.with_bbox:
            if self.with_mfm:
                self.ms_bbox_roi_extractor.init_weights()
                self.ss_bbox_roi_extractor.init_weights()
            else:
                self.bbox_roi_extractor.init_weights()
            # self.bbox_head.init_weights(pretrained)
            self.bbox_head.init_weights()
        #if self.with_mask:
        #    self.mask_head.init_weights()
        #    if not self.share_roi_extractor:
        #        self.mask_roi_extractor.init_weights()


    
    def _bbox_forward(self, x, rois):
        """Box head forward function used in both training and testing."""
        # TODO: a more flexible way to decide which feature maps to use
        if self.with_mfm:
            ss_bbox_feats = self.ss_bbox_roi_extractor(
                [x[-1]], rois)
            x = [self.mfm_fc(x[i]) for i in range(self.ms_bbox_roi_extractor.num_inputs)]
            ms_bbox_feats = self.ms_bbox_roi_extractor(
                x[:self.ms_bbox_roi_extractor.num_inputs], rois) # multi scale

            factor = self.mfm_factor.reshape(1, -1, 1, 1).expand_as(ms_bbox_feats)
            bbox_feats = ss_bbox_feats + ms_bbox_feats * factor
        else:
            bbox_feats = self.bbox_roi_extractor(
                x[:self.bbox_roi_extractor.num_inputs], rois)
        if self.with_shared_head:
            bbox_feats = self.shared_head(bbox_feats)

        if self.bbox_head.rois_mode is None:
            cls_score, bbox_pred = self.bbox_head(bbox_feats)                           # default
        else:
            cls_score, bbox_pred = self.bbox_head(bbox_feats, rois.clone().detach())    # `rois` does not require grad in bbox_head

        bbox_results = dict(
            cls_score=cls_score, bbox_pred=bbox_pred, bbox_feats=bbox_feats)
        return bbox_results




    


import json
import os.path as osp
import warnings

import numpy as np
from visionsuite.utils.dataset.formats.labelme.utils import get_polygon_points_from_labelme_shape

from visionsuite.engines.data.slicer.utils.polygons import get_intersected_points


def get_roi_image(roi, img_width, img_height):
    if len(roi) != 0:
        tl_x, tl_y, br_x, br_y = roi[0], roi[1], roi[2], roi[3]
    else:
        tl_x, tl_y, br_x, br_y = 0, 0, img_width, img_height

    return [tl_x, tl_y, br_x, br_y]


def get_center_point_from_points(points):
    xs, ys = [], []
    for point in points:
        xs.append(point[0])
        ys.append(point[1])

    cx = (np.max(xs) - np.min(xs)) / 2 + np.min(xs)
    cy = (np.max(ys) - np.min(ys)) / 2 + np.min(ys)

    return [cx, cy]


def finetune_patch_for_roi(cx, cy, patch_width, patch_height, br_x, br_y, logger=None):
    br_offset_x = cx + patch_width / 2 - br_x
    br_offset_y = cy + patch_height / 2 - br_y
    if br_offset_x > 0:
        cx -= br_offset_x
    if br_offset_y > 0:
        cy -= br_offset_y

    tl_offset_x = cx - patch_width / 2
    tl_offset_y = cy - patch_height / 2
    if tl_offset_x < 0:
        cx -= tl_offset_x
    if tl_offset_y < 0:
        cy -= tl_offset_y

    patch_coord = [
        int(cx - patch_width / 2),
        int(cy - patch_height / 2),
        int(cx + patch_width / 2),
        int(cy + patch_height / 2),
    ]
    
    patch_coord[0] -= (patch_width - (patch_coord[2] - patch_coord[0]))
    patch_coord[2] -= (patch_width - (patch_coord[2] - patch_coord[0]))
    patch_coord[1] -= (patch_height - (patch_coord[3] - patch_coord[1]))
    patch_coord[3] -= (patch_height - (patch_coord[3] - patch_coord[1]))

    if logger is not None:
        logger.assertion_log(
            patch_coord[2] - patch_coord[0] == patch_width
            and patch_coord[3] - patch_coord[1] == patch_height,
            RuntimeError(f"patch coord is wrong"),
            parent_fn=finetune_patch_for_roi.__name__,
        )

    return patch_coord


def get_annotations_from_labelme_json(img_file, logger=None):
    json_file = osp.splitext(img_file)[0] + ".json"
    # TODO: how to log?
    try:
        with open(json_file) as jf:
            anns = json.load(jf)
    except Exception as error:
        raise RuntimeError(
            f"There is something wrong within {json_file} to load: {error}"
        )

    return anns


def get_available_points_from_annotations(
    img_file, anns, mode, classes, roi, include_point_positive
):
    points = []
    for shape in anns["shapes"]:
        shape_type = str(shape["shape_type"]).lower()
        label = shape["label"].lower()

        if label in classes or label.upper() in classes:  # get all points
            _points = get_polygon_points_from_labelme_shape(
                shape, shape_type, include_point_positive, mode
            )
            if len(_points) == 0:  # there is no points
                continue

            if len(roi) != 0:  # get intersected points
                _points = get_points_in_roi(_points, roi, filename=img_file)
                if len(_points) == 0:  # when out of roi
                    continue

                for _point in _points:
                    points.append(
                        {"points": _point, "label": label, "shape_type": shape_type}
                    )
            else:
                points.append(
                    {"points": _points, "label": label, "shape_type": shape_type}
                )

    return points


def is_point_in_roi(point, roi):

    assert isinstance(point, list) and not isinstance(point[0], list), ValueError(
        f"point({point}) should be 1 dimension list"
    )
    assert len(point) == 2, ValueError(
        f"The length of points should be 2, not {len(point)}"
    )
    assert isinstance(roi, list) and not isinstance(roi[0], list), ValueError(
        f"roi({roi}) should be 1 dimension list"
    )
    assert len(roi) == 4, ValueError(f"The length of roi should be 4, not {len(roi)}")

    if (
        roi[0] <= point[0]
        and point[0] <= roi[2]
        and roi[1] <= point[1]
        and point[1] <= roi[3]
    ):
        return True
    else:
        return False


def get_points_in_roi(points, roi, filename=None):
    assert isinstance(points, list) and isinstance(points[0], list), ValueError(
        f"points({points}) should be 2 dimension list"
    )

    if len(points) <= 2:  # for a point and a line
        points_in_roi = []
        for point in points:
            if is_point_in_roi(point, roi):
                points_in_roi.append(point)

        if len(points_in_roi) != 0:
            return [points_in_roi]
        else:
            return []

    else:  # for a polygon
        is_fully_in = True
        for point in points:
            if not is_point_in_roi(point, roi):
                is_fully_in = False
                # If at least one of point is out of roi, we need to get intersection points.
                intersected_ponits = get_intersected_points(
                    roi,
                    points,
                    is_points1_roi=True,
                    output_list=True,
                    filename=filename,
                )

                if len(intersected_ponits) == 0:
                    return []
                else:
                    points = intersected_ponits
                    is_fully_in = False
                    warnings.warn(
                        f"Object is split, because some of points are out of RoI for {filename}"
                    )
                    break

        if is_fully_in:
            return [points]
        else:
            return points


def get_points_in_patch(
    points, roi, image_roi, include_point_positive=False, filename=None
):
    points_in_patch = []
    for points_dict in points:
        _points = points_dict["points"]

        if len(_points) > 2:  # for positive samples,
            _points = get_points_in_roi(_points, roi, filename=filename)
            for _point in _points:
                points_in_patch.append(
                    {
                        "points": _point,
                        "label": points_dict["label"],
                        "shape_type": points_dict["shape_type"],
                        "roi": image_roi,
                    }
                )

        elif (len(_points) > 0 and len(_points) <= 2):
            if include_point_positive:
                _points = get_points_in_roi(_points, roi, filename=filename)
                if len(_points) != 0:
                    points_in_patch.append({"points": [], 'label': "", 'shape_type': "", 'roi': image_roi})
            else:
                pass

        else:
            raise NotImplementedError(f"There is no such case for {_points} points")

    return points_in_patch


# def get_points_in_patches(points, patches_coord, image_roi, include_point_positive=False, filename=None):
#     points_in_patches = []
#     for patch_coord in patches_coord:
#         points_in_patch = get_points_in_patch(points, patch_coord, image_roi, \
#                             include_point_positive=include_point_positive, filename=filename)

#         points_in_patches.append(points_in_patch)

#     return points_in_patches



from visionsuite.engines.classification.utils.registry import SCHEDULERS
from visionsuite.engines.utils.helpers import get_params_from_obj


def build_scheduler(optimizer, epochs, iters_per_epoch, **config):
    
    warmup_total_iters = 0 
    if config['warmup_scheduler'] and 'total_iters' in config['warmup_scheduler'] and config['warmup_scheduler']['total_iters'] > 0:
        warmup_total_iters = config['warmup_scheduler']['total_iters']
    main_lr_scheduler = SCHEDULERS.get(config['type'], case_sensitive=config['case_sensitive'])(
        optimizer, 
        total_iters=iters_per_epoch * (epochs - warmup_total_iters), 
        power=config['power']
    )

    if config['warmup_scheduler'] and config['warmup_scheduler']['total_iters'] > 0:
        scheduler_obj = SCHEDULERS.get(config['warmup_scheduler']['type'], case_sensitive=config['warmup_scheduler']['case_sensitive'])
        if not scheduler_obj:
            raise RuntimeError(
                f"Invalid lr scheduler '{config['warmup_scheduler']['type']}'. Only ConstantLR and LinearLR are available")
        scheduler_params = get_params_from_obj(scheduler_obj)
        for key in scheduler_params.keys():
            if key in config['warmup_scheduler']:
                scheduler_params[key] = config['warmup_scheduler'][key]
                
            if key == 'optimizer':
                scheduler_params[key] = optimizer

        warmup_lr_scheduler = scheduler_obj(**scheduler_params)
        lr_scheduler = SCHEDULERS.get(config['warmup_scheduler']['integrate_scheduler']['type'],
                                      case_sensitive=config['warmup_scheduler']['integrate_scheduler']['case_sensitive'])(
            optimizer, schedulers=[warmup_lr_scheduler, main_lr_scheduler], milestones=[config['warmup_scheduler']['total_iters']]
        )

    else:
        lr_scheduler = main_lr_scheduler
        
    return lr_scheduler


def on_loop_build_start(loop, *args, **kwargs):
    pass

def on_loop_build_end(loop, *args, **kwargs):
    
    assert loop is not None, ValueError(f"loop is None")
    
    for attribute_name in loop.required_attributes:
        assert hasattr(loop, attribute_name), ValueError(f'{attribute_name} must be assgined in loop class')
        assert getattr(loop, attribute_name) is not None, ValueError(f"{attribute_name} is None for loop")
    

def on_loop_run_start(loop, *args, **kwargs):
    pass

def on_loop_run_end(loop, *args, **kwargs):
    pass

callbacks = {
    "on_loop_build_start": [on_loop_build_start],
    "on_loop_build_end": [on_loop_build_end],
    "on_loop_run_start": [on_loop_run_start],
    "on_loop_run_end": [on_loop_run_end],
}



import os.path as osp
from reportlab.lib.pagesizes import letter
from reportlab.pdfgen import canvas
from PIL import Image

def create_pdf(title, subtitle, output_dir):
    c = canvas.Canvas(osp.join(output_dir, 'analysis.pdf'), pagesize=letter)
    
    c.setFont("Helvetica-Bold", 24)
    c.drawCentredString(letter[0] / 2, letter[1] - 50, title)
    
    c.setFont("Helvetica", 18)
    c.drawCentredString(letter[0] / 2, letter[1] - 80, subtitle)
    
    ##########################################################################################
    try:
        content = "1. Number of Images in train/val dataset"
        c.drawText(get_text_object(c, content, 120))
        img_file = osp.join(output_dir, 'num_images_bar.png')
        img = Image.open(img_file)
        width, height = img.size
        ratio = height / float(width)
        aspect = 0.5
        c.drawInlineImage(img_file, 50, letter[1] - 350, width=letter[0]*aspect, height=letter[0]*ratio*aspect)

        img_file = osp.join(output_dir, 'num_images_pie.png')
        img = Image.open(img_file)
        width, height = img.size
        ratio = height / float(width)
        aspect = 0.4
        c.drawInlineImage(img_file, letter[0]/2 + 20, letter[1] - 380, width=letter[0]*aspect, height=letter[0]*ratio*aspect)
    except:
        pass
    
    ##########################################################################################
    try:
        content = "2. Number of Objects in train/val dataset"
        c.drawText(get_text_object(c, content, 410))
        img_file = osp.join(output_dir, 'train_pie_chart.png')
        img = Image.open(img_file)
        width, height = img.size
        ratio = height / float(width)
        aspect = 0.4
        c.drawInlineImage(img_file, 50, letter[1] - 700, width=letter[0]*aspect, height=letter[0]*ratio*aspect)

        img_file = osp.join(output_dir, 'val_pie_chart.png')
        img = Image.open(img_file)
        width, height = img.size
        ratio = height / float(width)
        aspect = 0.4
        c.drawInlineImage(img_file, letter[0]/2 + 20, letter[1] - 700, width=letter[0]*aspect, height=letter[0]*ratio*aspect)

        content = "The detailed information by each image is in the csv or html file."
        c.drawText(get_text_object(c, content, 710, 50))
    except:
        pass
    
    # ##########################################################################################
    try:
        content = "3-1. Size of Objects"
        c.drawText(get_text_object(c, content, 50))
        img_file = osp.join(output_dir, 'objects_size_by_labels.png')
        img = Image.open(img_file)
        width, height = img.size
        ratio = height / float(width)
        aspect = 0.7
        c.drawInlineImage(img_file, 50, letter[1] - 500, width=letter[0]*aspect, height=letter[0]*ratio*aspect)
    except: 
        pass

    c.save()


def get_text_object(c, content, loc_y, loc_x=50):
    c.setFont("Helvetica", 12)
    text_object = c.beginText(loc_x, letter[1] - loc_y)
    text_object.setFont("Helvetica", 12)
    
    for line in content.splitlines():
        text_object.textLine(line)
        
    return text_object



from __future__ import absolute_import

from keras_unet_collection.layer_utils import *
from keras_unet_collection.activations import GELU, Snake
from keras_unet_collection._backbone_zoo import backbone_zoo, bach_norm_checker

from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

def UNET_left(X, channel, kernel_size=3, stack_num=2, activation='ReLU', 
              pool=True, batch_norm=False, name='left0'):
    '''
    The encoder block of U-net.
    
    UNET_left(X, channel, kernel_size=3, stack_num=2, activation='ReLU', 
              pool=True, batch_norm=False, name='left0')
    
    Input
    ----------
        X: input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        stack_num: number of convolutional layers.
        activation: one of the `tensorflow.keras.layers` interface, e.g., 'ReLU'.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        batch_norm: True for batch normalization, False otherwise.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
        
    '''
    pool_size = 2
    
    X = encode_layer(X, channel, pool_size, pool, activation=activation, 
                     batch_norm=batch_norm, name='{}_encode'.format(name))

    X = CONV_stack(X, channel, kernel_size, stack_num=stack_num, activation=activation, 
                   batch_norm=batch_norm, name='{}_conv'.format(name))
    
    return X


def UNET_right(X, X_list, channel, kernel_size=3, 
               stack_num=2, activation='ReLU',
               unpool=True, batch_norm=False, concat=True, name='right0'):
    
    '''
    The decoder block of U-net.
    
    Input
    ----------
        X: input tensor.
        X_list: a list of other tensors that connected to the input tensor.
        channel: number of convolution filters.
        kernel_size: size of 2-d convolution kernels.
        stack_num: number of convolutional layers.
        activation: one of the `tensorflow.keras.layers` interface, e.g., 'ReLU'.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.
        batch_norm: True for batch normalization, False otherwise.
        concat: True for concatenating the corresponded X_list elements.
        name: prefix of the created keras layers.
        
    Output
    ----------
        X: output tensor.
    
    '''
    
    pool_size = 2
    
    X = decode_layer(X, channel, pool_size, unpool, 
                     activation=activation, batch_norm=batch_norm, name='{}_decode'.format(name))
    
    # linear convolutional layers before concatenation
    X = CONV_stack(X, channel, kernel_size, stack_num=1, activation=activation, 
                   batch_norm=batch_norm, name='{}_conv_before_concat'.format(name))
    if concat:
        # <--- *stacked convolutional can be applied here
        X = concatenate([X,]+X_list, axis=3, name=name+'_concat')
    
    # Stacked convolutions after concatenation 
    X = CONV_stack(X, channel, kernel_size, stack_num=stack_num, activation=activation, 
                   batch_norm=batch_norm, name=name+'_conv_after_concat')
    
    return X

def unet_2d_base(input_tensor, filter_num, stack_num_down=2, stack_num_up=2, 
                 activation='ReLU', batch_norm=False, pool=True, unpool=True, 
                 backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='unet'):
    
    '''
    The base of U-net with an optional ImageNet-trained backbone.
    
    unet_2d_base(input_tensor, filter_num, stack_num_down=2, stack_num_up=2, 
                 activation='ReLU', batch_norm=False, pool=True, unpool=True, 
                 backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='unet')
    
    ----------
    Ronneberger, O., Fischer, P. and Brox, T., 2015, October. U-net: Convolutional networks for biomedical image segmentation. 
    In International Conference on Medical image computing and computer-assisted intervention (pp. 234-241). Springer, Cham.
    
    Input
    ----------
        input_tensor: the input tensor of the base, e.g., `keras.layers.Inpyt((None, None, 3))`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of backbone options) ----------
        backbone_name: the bakcbone model name. Should be one of the `tensorflow.keras.applications` class.
                       None (default) means no backbone. 
                       Currently supported backbones are:
                       (1) VGG16, VGG19
                       (2) ResNet50, ResNet101, ResNet152
                       (3) ResNet50V2, ResNet101V2, ResNet152V2
                       (4) DenseNet121, DenseNet169, DenseNet201
                       (5) EfficientNetB[0-7]
        weights: one of None (random initialization), 'imagenet' (pre-training on ImageNet), 
                 or the path to the weights file to be loaded.
        freeze_backbone: True for a frozen backbone.
        freeze_batch_norm: False for not freezing batch normalization layers.
        
    Output
    ----------
        X: output tensor.
    
    '''
    
    activation_func = eval(activation)

    X_skip = []
    depth_ = len(filter_num)

    # no backbone cases
    if backbone is None:

        X = input_tensor

        # stacked conv2d before downsampling
        X = CONV_stack(X, filter_num[0], stack_num=stack_num_down, activation=activation, 
                       batch_norm=batch_norm, name='{}_down0'.format(name))
        X_skip.append(X)

        # downsampling blocks
        for i, f in enumerate(filter_num[1:]):
            X = UNET_left(X, f, stack_num=stack_num_down, activation=activation, pool=pool, 
                          batch_norm=batch_norm, name='{}_down{}'.format(name, i+1))        
            X_skip.append(X)

    # backbone cases
    else:
        # handling VGG16 and VGG19 separately
        if 'VGG' in backbone:
            backbone_ = backbone_zoo(backbone, weights, input_tensor, depth_, freeze_backbone, freeze_batch_norm)
            # collecting backbone feature maps
            X_skip = backbone_([input_tensor,])
            depth_encode = len(X_skip)
            
        # for other backbones
        else:
            backbone_ = backbone_zoo(backbone, weights, input_tensor, depth_-1, freeze_backbone, freeze_batch_norm)
            # collecting backbone feature maps
            X_skip = backbone_([input_tensor,])
            depth_encode = len(X_skip) + 1


        # extra conv2d blocks are applied
        # if downsampling levels of a backbone < user-specified downsampling levels
        if depth_encode < depth_:

            # begins at the deepest available tensor  
            X = X_skip[-1]

            # extra downsamplings
            for i in range(depth_-depth_encode):
                i_real = i + depth_encode

                X = UNET_left(X, filter_num[i_real], stack_num=stack_num_down, activation=activation, pool=pool, 
                              batch_norm=batch_norm, name='{}_down{}'.format(name, i_real+1))
                X_skip.append(X)

    # reverse indexing encoded feature maps
    X_skip = X_skip[::-1]
    # upsampling begins at the deepest available tensor
    X = X_skip[0]
    # other tensors are preserved for concatenation
    X_decode = X_skip[1:]
    depth_decode = len(X_decode)

    # reverse indexing filter numbers
    filter_num_decode = filter_num[:-1][::-1]

    # upsampling with concatenation
    for i in range(depth_decode):
        X = UNET_right(X, [X_decode[i],], filter_num_decode[i], stack_num=stack_num_up, activation=activation, 
                       unpool=unpool, batch_norm=batch_norm, name='{}_up{}'.format(name, i))

    # if tensors for concatenation is not enough
    # then use upsampling without concatenation 
    if depth_decode < depth_-1:
        for i in range(depth_-depth_decode-1):
            i_real = i + depth_decode
            X = UNET_right(X, None, filter_num_decode[i_real], stack_num=stack_num_up, activation=activation, 
                       unpool=unpool, batch_norm=batch_norm, concat=False, name='{}_up{}'.format(name, i_real))   
    return X

def unet_2d(input_size, filter_num, n_labels, stack_num_down=2, stack_num_up=2,
            activation='ReLU', output_activation='Softmax', batch_norm=False, pool=True, unpool=True, 
            backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='unet'):
    '''
    U-net with an optional ImageNet-trained bakcbone.
    
    unet_2d(input_size, filter_num, n_labels, stack_num_down=2, stack_num_up=2,
            activation='ReLU', output_activation='Softmax', batch_norm=False, pool=True, unpool=True, 
            backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='unet')
    
    ----------
    Ronneberger, O., Fischer, P. and Brox, T., 2015, October. U-net: Convolutional networks for biomedical image segmentation. 
    In International Conference on Medical image computing and computer-assisted intervention (pp. 234-241). Springer, Cham.
    
    Input
    ----------
        input_size: the size/shape of network input, e.g., `(128, 128, 3)`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        n_labels: number of output labels.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        output_activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interface or 'Sigmoid'.
                           Default option is 'Softmax'.
                           if None is received, then linear activation is applied.
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.                 
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of backbone options) ----------
        backbone_name: the bakcbone model name. Should be one of the `tensorflow.keras.applications` class.
                       None (default) means no backbone. 
                       Currently supported backbones are:
                       (1) VGG16, VGG19
                       (2) ResNet50, ResNet101, ResNet152
                       (3) ResNet50V2, ResNet101V2, ResNet152V2
                       (4) DenseNet121, DenseNet169, DenseNet201
                       (5) EfficientNetB[0-7]
        weights: one of None (random initialization), 'imagenet' (pre-training on ImageNet), 
                 or the path to the weights file to be loaded.
        freeze_backbone: True for a frozen backbone.
        freeze_batch_norm: False for not freezing batch normalization layers.
        
    Output
    ----------
        model: a keras model.
    
    '''
    activation_func = eval(activation)
    
    if backbone is not None:
        bach_norm_checker(backbone, batch_norm)
        
    IN = Input(input_size)
    
    # base    
    X = unet_2d_base(IN, filter_num, stack_num_down=stack_num_down, stack_num_up=stack_num_up, 
                     activation=activation, batch_norm=batch_norm, pool=pool, unpool=unpool, 
                     backbone=backbone, weights=weights, freeze_backbone=freeze_backbone, 
                     freeze_batch_norm=freeze_backbone, name=name)
    
    # output layer
    OUT = CONV_output(X, n_labels, kernel_size=1, activation=output_activation, name='{}_output'.format(name))
    
    # functional API model
    model = Model(inputs=[IN,], outputs=[OUT,], name='{}_model'.format(name))
    
    return model



from visionsuite.engines.segmentation.utils.registry import MODELS

def build_model(**config):
    model = MODELS.get(config['type'], 
                       case_sensitive=config['case_sensitive'])()
    
    return model



# import torchvision 
# import os.path as osp 
# import torch 
# import torch.nn as nn 

# def get_model(args, num_classes, device):
        
#     if args['model'] == 'unet3plus':        
#         from models.unet3plus.models.UNet_3Plus import UNet_3Plus
#         model = UNet_3Plus(n_classes=num_classes)
        
#         model.to(device)
#         if args['distributed']:
#             model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)

#         model_without_ddp = model
#         if args['distributed']:
#             model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args['gpu']])
#             model_without_ddp = model.module

#         params_to_optimize = [
#             {"params": [p for p in model_without_ddp.parameters() if p.requires_grad]},
#         ]
#     else:
#         model = torchvision.models.get_model(
#                 args['model'],
#                 weights=args['weights'],
#                 weights_backbone=args['weights_backbone'],
#                 num_classes=num_classes,
#                 aux_loss=args['aux_loss'],
#         )
        
#         model.to(device)
#         if args['distributed'] and args['distributed']['use']:
#             model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args['distributed']['gpu']])
#             model_without_ddp = model.module
#         else:
#             if torch.cuda.device_count() > 1 and len(args['train']['device_ids']) > 1:
#                 model = nn.DataParallel(model, device_ids=args['train']['device_ids'], 
#                                         output_device=args['train']['device_ids'][0])
#                 model_without_ddp = model.module
#             else:
#                 model_without_ddp = model
            
#         params_to_optimize = [
#             {"params": [p for p in model_without_ddp.backbone.parameters() if p.requires_grad]},
#             {"params": [p for p in model_without_ddp.classifier.parameters() if p.requires_grad]},
#         ]
#         if args['aux_loss']:
#             params = [p for p in model_without_ddp.aux_classifier.parameters() if p.requires_grad]
#             params_to_optimize.append({"params": params, "lr": args['lr'] * 10})
    
#     return model, model_without_ddp, params_to_optimize


#!/usr/bin/env python3

from .accuracy import calc_metrics
from .binary_edge_loss import BinaryEdgeLoss, ConsensusBinaryEdgeLoss
from .multilabel_edge_loss import (
    MultiLabelEdgeLoss,
    BalancedMultiLabelLoss,
    WeightedMultiLabelLoss,
)

__all__ = [
    "calc_metrics",
    "BinaryEdgeLoss",
    "ConsensusBinaryEdgeLoss",
    "MultiLabelEdgeLoss",
    "BalancedMultiLabelLoss",
    "WeightedMultiLabelLoss",
]



from visionsuite.engines.data.slicer.src.background_patches import get_background_patches
from visionsuite.engines.data.slicer.src.centric_patches import get_centric_patches
from visionsuite.engines.data.slicer.src.sliding_patches import get_sliding_patches
from visionsuite.engines.data.slicer.utils.functional import (
    get_annotations_from_labelme_json,
    get_available_points_from_annotations,
)
from visionsuite.engines.data.slicer.utils.polygons import handle_self_intersection


# TODO: Need to add logger
def get_imgs_info_from_patches(
    mode, img_file, classes, patch_info, roi=[], logger=None
):

    if logger is not None:
        logger.assertion_log(
            patch_info["sliding"] or patch_info["centric"],
            RuntimeError(
                f"If you want to use patch, need to choose at least one of sliding or centric"
            ),
            parent_fn=get_imgs_info_from_patches.__name__,
        )

    anns = get_annotations_from_labelme_json(img_file, logger=logger)
    img_width, img_height = anns["imageWidth"], anns["imageHeight"]
    num_data = 0
    rois = []
    background_patches_rois, background_patches_num_data = [], 0
    labels = []

    if len(anns["shapes"]) != 0:  # for positive samples

        # get candidate patch coordinates ---
        points = get_available_points_from_annotations(
            img_file, anns, mode, classes, roi, patch_info["include_point_positive"]
        )

        points = handle_self_intersection(points)

        # background patches ---
        # if patch_info['bg_ratio_by_image'] > 0:
        background_patches_rois, background_patches_num_data = get_background_patches(
            img_height=img_height,
            img_width=img_width,
            patch_height=patch_info['height'],
            patch_width=patch_info['width'],
            points=points,
            overlap_ratio=patch_info['overlap_ratio'],
            roi=roi,
            skip_highly_overlapped_tiles=False)

        if len(points) > 0:
            # centric patches ---
            if patch_info["centric"]:
                centric_patches_rois, centric_patches_num_data, rois_labels = (
                    get_centric_patches(
                        points,
                        patch_info["width"],
                        patch_info["height"],
                        img_width,
                        img_height,
                        roi=roi,
                        include_point_positive=patch_info["include_point_positive"],
                        filename=img_file,
                        logger=logger,
                    )
                )
                rois += centric_patches_rois
                num_data += centric_patches_num_data
                labels += rois_labels

                if logger is not None:
                    logger.assertion_log(
                        (len(rois) == num_data) and (len(labels) == num_data),
                        RuntimeError(
                            f"The number rois ({len(rois)}) should be same to the number of labels ({len(labels)})"
                        ),
                        parent_fn=get_imgs_info_from_patches.__name__,
                    )

            # sliding patches ---
            if patch_info["sliding"]:
                overlap_ratio = patch_info["overlap_ratio"]
                num_involved_pixel = patch_info["num_involved_pixel"]
                bg_ratio = patch_info["sliding_bg_ratio"]

                patch_coords, num_patch_sliding, rois_labels = get_sliding_patches(
                    img_height=img_height,
                    img_width=img_width,
                    patch_height=patch_info["height"],
                    patch_width=patch_info["width"],
                    points=points,
                    overlap_ratio=overlap_ratio,
                    num_involved_pixel=num_involved_pixel,
                    bg_ratio=bg_ratio,
                    roi=roi,
                    include_point_positive=patch_info["include_point_positive"],
                    skip_highly_overlapped_tiles=(mode != "train"),
                    filename=img_file,
                )
                # for patch_coord in patch_coords:
                #     try:
                #         assert patch_coord[2] - patch_coord[0] == patch_info['width'] and patch_coord[3] - patch_coord[
                #             1] == patch_info['height'], RuntimeError(f"patch coord ({patch_coord}) is wrong")
                #     except AssertionError as e:
                #         if logger is not None:
                #             logger.error(e, get_imgs_info_from_patches.__name__)
                #         raise e

                #     rois.append(patch_coord)
                rois += patch_coords
                num_data += num_patch_sliding
                labels += rois_labels

    # TODO: Need to consider empty json by using sliding patch?
    # else: # for negative samples
    #     if patch_info['sliding']:
    #         overlap_ratio=patch_info['overlap_ratio']
    #         num_involved_pixel=patch_info['num_involved_pixel']
    #         bg_ratio=patch_info['sliding_bg_ratio']
    #     else:
    #         overlap_ratio = 0.1
    #         num_involved_pixel = 10
    #         bg_ratio = 0

    #     patch_coords, num_patch_sliding, rois_labels = get_sliding_patches(img_height=img_height, img_width=img_width, \
    #                         patch_height=patch_info['height'], patch_width=patch_info['width'], points=[], \
    #                         overlap_ratio=overlap_ratio, num_involved_pixel=num_involved_pixel, \
    #                         bg_ratio=bg_ratio, roi=roi, include_point_positive=patch_info['include_point_positive'], \
    #                         skip_highly_overlapped_tiles=(mode is not 'train'), filename=img_file)

    #     for patch_coord, roi_labels in zip(patch_coords, rois_labels):
    #         try:
    #             assert patch_coord[2] - patch_coord[0] == patch_info['width'] and patch_coord[3] - patch_coord[1] == patch_info['height'], \
    #                     f"patch coord is wrong"
    #         except AssertionError as e:
    #             if logger is not None:
    #                 logger.error(e, get_imgs_info_from_patches.__name__)
    #             raise e

    #         rois.append(patch_coord)
    #         labels.append([roi_labels])
    #     num_data += num_patch_sliding

    return rois, num_data, background_patches_rois, labels



import datetime
import json
import os
import os.path as osp

# import socket

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


def get_config(config_json, log_stream_level, log_file_level, log_dir, name=None):
    with open(config_json) as f:
        config = json.load(f)

    set_log_level(config, log_stream_level, log_file_level)
    set_log_dir(config, log_dir, name)

    return config


def set_log_level(config, log_stream_level, log_file_level):
    log_stream_level = log_stream_level.upper()
    log_file_level = log_file_level.upper()
    assert log_stream_level.upper() in get_log_levels(), ValueError(
        f"Log-level for streaming should be one of {LOG_LEVELS}"
    )
    assert log_file_level.upper() in get_log_levels(), ValueError(
        f"Log-level for file should be one of {LOG_LEVELS}"
    )

    config["handlers"]["stream"]["level"] = log_stream_level
    config["root"]["level"] = log_file_level


def set_log_dir(config, log_dir, name=None):
    if log_dir is None:
        # hostname = socket.gethostname()

        # log_dir = f'/home/{hostname}/logs'
        log_dir = "/DeepLearning/_logs/logger"
        current_date = datetime.datetime.now()
        year = current_date.year
        month = current_date.month
        day = current_date.day
        hour = current_date.hour

        log_dir = osp.join(log_dir, f"{year}_{month}_{day}", str(hour))

        if not osp.exists(log_dir):
            os.makedirs(log_dir)

    assert log_dir is not None, ValueError(
        f"Log directory should be assigned, not {log_dir}"
    )

    if name is None:
        config["handlers"]["file"]["filename"] = osp.join(
            log_dir, config["handlers"]["file"]["filename"]
        )
    else:
        config["handlers"]["file"]["filename"] = osp.join(log_dir, name + ".log")


def get_log_levels():
    return LOG_LEVELS


# Copyright (c) OpenMMLab. All rights reserved.
from mmrotate.models.roi_heads.bbox_heads.convfc_rbbox_head import (RotatedConvFCBBoxHead,
                                RotatedKFIoUShared2FCBBoxHead,
                                RotatedShared2FCBBoxHead)
from mmrotate.models.roi_heads.bbox_heads.gv_bbox_head import GVBBoxHead
from mmrotate.models.roi_heads.bbox_heads.rotated_bbox_head import RotatedBBoxHead

from .rotated_mae_bbox_head import RotatedMAEBBoxHead
from .rotated_mae_bbox_head_stdc import RotatedMAEBBoxHeadSTDC


__all__ = [
    'RotatedBBoxHead', 'RotatedConvFCBBoxHead', 'RotatedShared2FCBBoxHead',
    'GVBBoxHead', 'RotatedKFIoUShared2FCBBoxHead',

    'RotatedMAEBBoxHead', 'RotatedMAEBBoxHeadSTDC',
]



import argparse 
from types import SimpleNamespace
from visionsuite.engines.utils.loggers import Logger

class BaseOOPModule(Logger):
    def __init__(self, name=None):
        super().__init__(name)
        self._status = {}
        
    def track_status(func):
        def wrapper(self, *args, **kwargs):
            # Update the status dictionary for the function being executed
            self._status[func.__name__] = True
            return func(self, *args, **kwargs)
        return wrapper

    @property 
    def status(self):
        return self._status 
    
    @track_status
    def build(self, *args, **kwargs):
        print(f"args: ", args)
        print(f"kwargs: ", kwargs)
        
        if '_logger' in kwargs and kwargs['_logger'] and 'logs_dir' in kwargs['_logger']:
            self.set_logger(log_stream_level=kwargs['_logger'].get('log_stream_level', 'DEBUG'),
                        log_file_level=kwargs['_logger'].get('log_file_level', 'DEBUG'),
                        log_dir=kwargs['_logger']['logs_dir']) 
            
        
        if isinstance(kwargs, (argparse.Namespace, SimpleNamespace)):
            self.args = dict(kwargs)
        elif isinstance(kwargs, dict):
            self.args = kwargs
        else:
            NotImplementedError(f"NOT Considered this case for args({args}) and kwargs({kwargs})")
        
        assert self.args is not None, RuntimeError(f"Args for dataset is None")
        self.log_info(f"Loaded args: {self.args}", self.build.__name__, __class__.__name__)


    # if args.test_only:
    #     # We disable the cudnn benchmarking because it can noticeably affect the accuracy
    #     torch.backends.cudnn.benchmark = False
    #     torch.backends.cudnn.deterministic = True
    #     if model_ema:
    #         evaluate(model_ema, criterion, data_loader_test, device=device, log_suffix="EMA")
    #     else:
    #         evaluate(model, criterion, data_loader_test, device=device)
    #     return


#!/usr/bin/env python3

import argparse
import copy
import os
import os.path as osp
import time
import warnings

import torch
import torch.distributed as dist
import os

import mmcv
# from mmcv.cnn.utils import revert_sync_batchnorm
from mmcv.runner import get_dist_info, init_dist
from mmcv.utils import Config, DictAction, get_git_hash

from bbseg import __version__
from mmseg.apis import init_random_seed, set_random_seed, train_segmentor
from bbseg.datasets import build_dataset
from bbseg.models import build_segmentor
from mmseg.utils import collect_env, get_root_logger, setup_multi_processes, get_device

# torch.autograd.set_detect_anomaly(True)


def parse_args():
    parser = argparse.ArgumentParser(description="Train a segmentor")
    # parser.add_argument("--config", default='/HDD/_projects/github/VisionSuite/visionsuite/cores/sbcb/configs/deeplabv3plus/deeplabv3plus_r50b-d8_512x1024_40k_cityscapes.py')
    parser.add_argument("--config", default='/HDD/_projects/github/VisionSuite/visionsuite/cores/sbcb/configs/deeplabv3plus/deeplabv3plus_r101b-d8_1024_1024_lx.py')
    # parser.add_argument("--config", default='/HDD/_projects/github/VisionSuite/visionsuite/cores/sbcb/configs/deeplabv3plus/deeplabv3plus_r101b-d8_512_512_mask.py')
    parser.add_argument("--work-dir", default='/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/outputs/mm/train/sbcb')
    parser.add_argument("--load-from", help="the checkpoint file to load weights from")
    parser.add_argument("--resume-from", help="the checkpoint file to resume from")
    parser.add_argument(
        "--no-validate",
        action="store_true",
        help="whether not to evaluate the checkpoint during training",
    )
    group_gpus = parser.add_mutually_exclusive_group()
    group_gpus.add_argument(
        "--gpus",
        type=int,
        help="(Deprecated, please use --gpu-id) number of gpus to use "
        "(only applicable to non-distributed training)",
    )
    group_gpus.add_argument(
        "--gpu-ids",
        type=int,
        nargs="+",
        help="(Deprecated, please use --gpu-id) ids of gpus to use "
        "(only applicable to non-distributed training)",
    )
    group_gpus.add_argument(
        "--gpu-id",
        type=int,
        help="id of gpu to use " "(only applicable to non-distributed training)",
    )
    parser.add_argument("--seed", type=int, default=None, help="random seed")
    parser.add_argument(
        "--diff_seed",
        action="store_true",
        help="Whether or not set different seeds for different ranks",
    )
    parser.add_argument(
        "--deterministic",
        action="store_true",
        help="whether to set deterministic options for CUDNN backend.",
    )
    parser.add_argument(
        "--options",
        nargs="+",
        action=DictAction,
        help="--options is deprecated in favor of --cfg_options' and it will "
        "not be supported in version v0.22.0. Override some settings in the "
        "used config, the key-value pair in xxx=yyy format will be merged "
        "into config file. If the value to be overwritten is a list, it "
        'should be like key="[a,b]" or key=a,b It also allows nested '
        'list/tuple values, e.g. key="[(a,b),(c,d)]" Note that the quotation '
        "marks are necessary and that no white space is allowed.",
    )
    parser.add_argument(
        "--cfg-options",
        nargs="+",
        action=DictAction,
        help="override some settings in the used config, the key-value pair "
        "in xxx=yyy format will be merged into config file. If the value to "
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        "Note that the quotation marks are necessary and that no white space "
        "is allowed.",
    )
    parser.add_argument(
        "--launcher",
        choices=["none", "pytorch", "slurm", "mpi"],
        default="pytorch",
        help="job launcher",
    )
    parser.add_argument("--local_rank", type=int, default=0)
    parser.add_argument(
        "--auto-resume",
        action="store_true",
        help="resume from the latest checkpoint automatically.",
    )
    args = parser.parse_args()
    if "LOCAL_RANK" not in os.environ:
        os.environ["LOCAL_RANK"] = str(args.local_rank)

    if args.options and args.cfg_options:
        raise ValueError(
            "--options and --cfg-options cannot be both "
            "specified, --options is deprecated in favor of --cfg-options. "
            "--options will not be supported in version v0.22.0."
        )
    if args.options:
        warnings.warn(
            "--options is deprecated in favor of --cfg-options. "
            "--options will not be supported in version v0.22.0."
        )
        args.cfg_options = args.options

    return args


def main():
    args = parse_args()

    cfg = Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    # set cudnn_benchmark
    if cfg.get("cudnn_benchmark", False):
        torch.backends.cudnn.benchmark = True

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get("work_dir", None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join(
            "./work_dirs", osp.splitext(osp.basename(args.config))[0]
        )
    if args.load_from is not None:
        cfg.load_from = args.load_from
    if args.resume_from is not None:
        cfg.resume_from = args.resume_from

    if args.gpus is not None:
        cfg.gpu_ids = range(1)
        warnings.warn(
            "`--gpus` is deprecated because we only support "
            "single GPU mode in non-distributed training. "
            "Use `gpus=1` now."
        )
    if args.gpu_ids is not None:
        cfg.gpu_ids = args.gpu_ids[0:1]
        warnings.warn(
            "`--gpu-ids` is deprecated, please use `--gpu-id`. "
            "Because we only support single GPU mode in "
            "non-distributed training. Use the first GPU "
            "in `gpu_ids` now."
        )
    if args.gpus is None and args.gpu_ids is None:
        cfg.gpu_ids = [args.gpu_id]

    cfg.auto_resume = args.auto_resume

    # init distributed env first, since logger depends on the dist info.
    if args.launcher == "none":
        distributed = False
    else:
        distributed = True
        init_dist(args.launcher, **cfg.dist_params)
        # gpu_ids is used to calculate iter when resuming checkpoint
        _, world_size = get_dist_info()
        cfg.gpu_ids = range(world_size)

    # create work_dir
    mmcv.mkdir_or_exist(osp.abspath(cfg.work_dir))
    # dump config
    cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
    # init the logger before other steps
    timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())
    log_file = osp.join(cfg.work_dir, f"{timestamp}.log")
    logger = get_root_logger(log_file=log_file, log_level=cfg.log_level)

    # set multi-process settings
    setup_multi_processes(cfg)

    # init the meta dict to record some important information such as
    # environment info and seed, which will be logged
    meta = dict()
    # log env info
    env_info_dict = collect_env()
    env_info = "\n".join([f"{k}: {v}" for k, v in env_info_dict.items()])
    dash_line = "-" * 60 + "\n"
    logger.info("Environment info:\n" + dash_line + env_info + "\n" + dash_line)
    meta["env_info"] = env_info

    # log some basic info
    logger.info(f"Distributed training: {distributed}")
    logger.info(f"Config:\n{cfg.pretty_text}")

    # set random seeds
    # NOTE: should use seed of 1111 and deterministic mode
    cfg.device = get_device()
    seed = init_random_seed(args.seed, device=cfg.device)
    seed = seed + dist.get_rank() if args.diff_seed else seed
    logger.info(f"Set random seed to {seed}, " f"deterministic: {args.deterministic}")
    set_random_seed(seed, deterministic=args.deterministic)
    cfg.seed = seed
    meta["seed"] = seed
    meta["exp_name"] = osp.basename(args.config)

    model = build_segmentor(
        cfg.model, train_cfg=cfg.get("train_cfg"), test_cfg=cfg.get("test_cfg")
    )
    model.init_weights()

    # # SyncBN is not support for DP
    # if not distributed:
    #     warnings.warn(
    #         "SyncBN is only supported with DDP. To be compatible with DP, "
    #         "we convert SyncBN to BN. Please use dist_train.sh which can "
    #         "avoid this error."
    #     )
    #     model = revert_sync_batchnorm(model)

    logger.info(model)

    datasets = [build_dataset(cfg.data.train)]
    
    if len(cfg.workflow) == 2:
        val_dataset = copy.deepcopy(cfg.data.val)
        val_dataset.pipeline = cfg.data.train.pipeline
        datasets.append(build_dataset(val_dataset))
    if cfg.checkpoint_config is not None:
        # save bbseg version, config file content and class names in
        # checkpoints as meta data
        cfg.checkpoint_config.meta = dict(
            bbseg_version=f"{__version__}+{get_git_hash()[:7]}",
            config=cfg.pretty_text,
            CLASSES=datasets[0].CLASSES,
            PALETTE=datasets[0].PALETTE,
        )
    # add an attribute for visualization convenience
    model.CLASSES = datasets[0].CLASSES
    # passing checkpoint meta for saving best checkpoint
    meta.update(cfg.checkpoint_config.meta)
    torch.cuda.empty_cache()
    torch.autograd.set_detect_anomaly(True)

    train_segmentor(
        model,
        datasets,
        cfg,
        distributed=distributed,
        validate=(not args.no_validate),
        timestamp=timestamp,
        meta=meta,
    )


if __name__ == "__main__":
    main()



import torch
import warnings
from abc import abstractmethod

from visionsuite.engines.utils.metrics.metric_logger import MetricLogger
from visionsuite.engines.classification.utils.metrics.accuracy import get_accuracies
from visionsuite.engines.utils.torch_utils.dist import reduce_across_processes
from visionsuite.engines.classification.utils.registry import VALIDATORS
from visionsuite.engines.utils.bases import BaseOOPModule
from visionsuite.engines.utils.callbacks import Callbacks
from visionsuite.engines.classification.utils.results import ValResults
from .callbacks import callbacks


@VALIDATORS.register()
class BaseValidator(BaseOOPModule, Callbacks):
    
    required_attributes = ['model', 'dataloader']
    
    def __init__(self):
        BaseOOPModule.__init__(self, name="BaseValidator")
        Callbacks.__init__(self)
        
        self.add_callbacks(callbacks)
        
    def build(self, model, loss, dataloader, device, label2index, archive=None, print_freq=100, topk=3, **args):
        super().build(**args)
        self.run_callbacks('on_validator_build_start')
        self.epoch = None
        self.model = model.model_ema if model.model_ema else model.model
        self.loss = loss
        self.dataloader = dataloader
        self.label2index = label2index
        self.args = args
        self.log_suffix = "EMA" if model.model_ema else ""
        self.archive = archive
        self.device = device
        self.print_freq = print_freq
        self.topk = topk
        
        self.results = ValResults()
        self.log_info(f"SET ValResults", self.build.__name__, __class__.__name__)
        
        self.metric_logger = MetricLogger(delimiter="  ")
        self.log_info(f"SET MetricLogger", self.build.__name__, __class__.__name__)
        
        self.run_callbacks('on_validator_build_end')
        
    @abstractmethod
    def run(self, epoch):
        self.epoch = epoch
        if epoch%self.args['epoch'] == 0:
            self.model.eval()
            header = f"Val.: {self.log_suffix}"
            num_processed_samples = 0
            start_time_epoch = 0
            self.run_callbacks('on_validator_epoch_start')
            self.log_info(f"START val epoch: {epoch}", self.run.__name__, __class__.__name__)
            with torch.inference_mode():
                for image, target in self.metric_logger.log_every(self.dataloader, self.print_freq, header):
                    self.run_callbacks('on_validator_batch_start')
                    image = image.to(self.device, non_blocking=True)
                    target = target.to(self.device, non_blocking=True)
                    self.log_debug(f"- image: {image.shape} with device({self.device})", self.run.__name__, __class__.__name__)
                    self.log_debug(f"- target: {target.shape} with device({self.device})", self.run.__name__, __class__.__name__)
                    output = self.model(image)

                    self._update_logger(output, target, batch_size=image.shape[0])
                    num_processed_samples += image.shape[0]
                    self.run_callbacks('on_validator_batch_start')
                    
            # gather the stats from all processes
            num_processed_samples = reduce_across_processes(num_processed_samples)
            if (
                hasattr(self.dataloader.dataset, "__len__")
                and len(self.dataloader.dataset) != num_processed_samples
                and torch.distributed.get_rank() == 0
            ):
                # See FIXME above
                warnings.warn(
                    f"It looks like the dataset has {len(self.dataloader.dataset)} samples, but {num_processed_samples} "
                    "samples were used for the validation, which might bias the results. "
                    "Try adjusting the batch size and / or the world size. "
                    "Setting the world size to 1 is always a safe bet."
                )
            # print(f"{header} Acc@1 {self.metric_logger.acc1.global_avg:.3f} Acc@5 {self.metric_logger.acc5.global_avg:.3f}")
            self.run_callbacks('on_validator_epoch_end', epoch=epoch, 
                               start_time_epoch=start_time_epoch)

    def _update_logger(self, output, target, batch_size):
        if self.metric_logger is not None:
            loss = self.loss(output, target)
            acc1, acc5 = get_accuracies(output, target, topk=(1, self.topk))
            # FIXME need to take into account that the datasets
            # could have been padded in distributed setup
            self.metric_logger.update(loss=loss.item())
            self.metric_logger.meters["acc1"].update(acc1.item(), n=batch_size)
            self.metric_logger.meters["acc5"].update(acc5.item(), n=batch_size)
            
            self.log_debug(f"- Updated metric_logger", self._update_logger.__name__, __class__.__name__)
            
            


#!/usr/bin/env python3

import warnings

import mmcv
import mmseg
import pyEdgeEval
from packaging.version import parse

from .version import __version__, version_info

"""Version managing for dependencies:
Depends on:
- ``mmcv``
- ``mmsegmentation``
- ``pyEdgeEval``
"""

MMCV_MIN = "1.3.13"
MMCV_MAX = "1.7.1"

MMSEG_MIN = "0.24.0"
MMSEG_MAX = "0.30.0"

EDGEEVAL_MIN = "0.2.0"
EDGEEVAL_MAX = "0.3.0"


def digit_version(version_str: str, length: int = 4):
    """Convert a version string into a tuple of integers.

    This method is usually used for comparing two versions. For pre-release
    versions: alpha < beta < rc.

    Args:
        version_str (str): The version string.
        length (int): The maximum number of version levels. Default: 4.

    Returns:
        tuple[int]: The version info in digits (integers).
    """
    version = parse(version_str)
    assert version.release, f"failed to parse version {version_str}"
    release = list(version.release)
    release = release[:length]
    if len(release) < length:
        release = release + [0] * (length - len(release))
    if version.is_prerelease:
        mapping = {"a": -3, "b": -2, "rc": -1}
        val = -4
        # version.pre can be None
        if version.pre:
            if version.pre[0] not in mapping:
                warnings.warn(
                    f"unknown prerelease version {version.pre[0]}, "
                    "version checking may go wrong"
                )
            else:
                val = mapping[version.pre[0]]
            release.extend([val, version.pre[-1]])
        else:
            release.extend([val, 0])

    elif version.is_postrelease:
        release.extend([1, version.post])
    else:
        release.extend([0, 0])
    return tuple(release)


mmcv_min_version = digit_version(MMCV_MIN)
mmcv_max_version = digit_version(MMCV_MAX)
mmcv_version = digit_version(mmcv.__version__)


assert mmcv_min_version <= mmcv_version <= mmcv_max_version, (
    f"MMCV=={mmcv.__version__} is used but incompatible. "
    f"Please install mmcv>={mmcv_min_version}, <={mmcv_max_version}."
)

mmseg_min_version = digit_version(MMSEG_MIN)
mmseg_max_version = digit_version(MMSEG_MAX)
mmseg_version = digit_version(mmseg.__version__)

assert mmseg_min_version <= mmseg_version < mmseg_max_version, (
    f"MMSeg=={mmseg.__version__} is used but incompatible. "
    f"Please install mmseg>={MMSEG_MIN}, <{MMSEG_MAX}."
)

edgeeval_min_version = digit_version(EDGEEVAL_MIN)
edgeeval_max_version = digit_version(EDGEEVAL_MAX)
edgeeval_version = digit_version(pyEdgeEval.__version__)

assert edgeeval_min_version <= edgeeval_version < edgeeval_max_version, (
    f"pyEdgeEval=={pyEdgeEval.__version__} is used but incompatible. "
    f"Please install pyEdgeEval>={EDGEEVAL_MIN}, <{EDGEEVAL_MAX}."
)

__all__ = ["__version__", "version_info", "digit_version"]



from visionsuite.engines.data.slicer.utils.functional import get_intersected_points, get_points_in_patch


def get_sliding_patches(
    img_width,
    img_height,
    patch_height,
    patch_width,
    points,
    overlap_ratio=0.2,
    num_involved_pixel=2,
    bg_ratio=-1,
    roi=[],
    include_point_positive=False,
    skip_highly_overlapped_tiles=False,
    filename=None,
):

    # TODO: Need to consider num_involved_pixel !!!!!!!!!
    # TODO: ignore the bg_ratio b/c there is already function for background_patches

    if len(roi) != 0 and roi is not None:
        tl_x, tl_y, br_x, br_y = roi[0], roi[1], roi[2], roi[3]
    else:
        tl_x, tl_y, br_x, br_y = 0, 0, img_width, img_height

    dx = int((1.0 - overlap_ratio) * patch_width)
    dy = int((1.0 - overlap_ratio) * patch_height)

    sliding_patches_rois = []
    sliding_patches_num_data = 0
    labels_in_patches = []
    for y0 in range(tl_y, br_y, dy):
        for x0 in range(tl_x, br_x, dx):
            if y0 + patch_height > br_y:
                if skip_highly_overlapped_tiles:  # skip if too much overlap (> 0.6)
                    if (y0 + patch_height - br_y) > (0.6 * patch_height):
                        continue
                    else:
                        y = br_y - patch_height
                else:
                    y = br_y - patch_height
            else:
                y = y0

            if x0 + patch_width > br_x:
                if skip_highly_overlapped_tiles:  # skip if too much overlap (> 0.6)
                    if (x0 + patch_width - br_x) > (0.6 * patch_width):
                        continue
                    else:
                        x = br_x - patch_width
                else:
                    x = br_x - patch_width
            else:
                x = x0

            xmin, xmax, ymin, ymax = x, x + patch_width, y, y + patch_height

            # when more than one point lies on the window, the window is considered patch
            is_part_in = []
            areas = []
            area = 0
            if len(points) > 0:
                for b in points:
                    for xb0, yb0 in b["points"]:
                        if (xb0 >= xmin) and (xb0 <= xmax) and (yb0 <= ymax) and (yb0 >= ymin):
                            is_part_in.append(1)
                        else:
                            is_part_in.append(0)

                    if all(is_part_in) or not all(is_part_in) and sum(is_part_in) != 0:
                        if len(b["points"]) > 2:
                            intersection_points, area = get_intersected_points(
                                [x, y, x + patch_width, y + patch_height],
                                b["points"],
                                is_points1_roi=True,
                                output_list=True,
                                get_area=True,
                                filename=filename,
                            )
                            areas.append(area)
                        elif 0 < len(b["points"]) and len(b["points"]) <= 2:
                            is_part_in.append(1)
                        else:
                            NotImplementedError(f"There is no such case considered When b['points'] is {b['points']}")

            flag_area_satisfied = False
            for _area in areas:
                if _area > num_involved_pixel:
                    flag_area_satisfied = True
                    break

            if not flag_area_satisfied:
                if sum(areas) > num_involved_pixel:
                    flag_area_satisfied = True

            if len(areas) == 0:
                flag_area_satisfied = True

            if sum(is_part_in) == 0 or not flag_area_satisfied:
                continue

            patch_coord = [x, y, x + patch_width, y + patch_height]
            points_in_patch = get_points_in_patch(
                points,
                patch_coord,
                [tl_x, tl_y, br_x, br_y],
                include_point_positive=include_point_positive,
                filename=filename,
            )
            if len(points_in_patch) != 0:
                labels_in_patches.append(points_in_patch)
                sliding_patches_rois.append(patch_coord)
                sliding_patches_num_data += 1

    return sliding_patches_rois, sliding_patches_num_data, labels_in_patches



try:
    from ultralytics import YOLO
except (ImportError, AssertionError):
    import subprocess
    subprocess.run(["pip", "install", "ultralytics"])
    from ultralytics import YOLO
from visionsuite.cores.roboflow.utils.parsing import get_params

def export(weights: str, target: str, params):
    
    reload_onnx = False
    params = get_params(params)
    if 'input_names' in params and isinstance(params['input_names'], str):
        params['input_names'] = [name.rstrip().lstrip() for name in params['input_names'].split(',')]
        reload_onnx = True
    if 'output_names' in params and isinstance(params['output_names'], str):
        params['output_names'] = [name.rstrip().lstrip() for name in params['output_names'].split(',')]
        reload_onnx = True
        
    model = YOLO(weights)
    print(f"Successfully LOADED weights: {weights} ")
    
    if target == 'onnx':
        export2onnx(model, weights, reload_onnx)
    else:
        NotImplementedError(f'There is no such case for {target}')

def export2onnx(model, weights, reload_onnx=False):
    import onnx
    model.export(format='onnx', opset=14)
    onnx_path = weights.replace(".pt", '.onnx')
    print(f"Saved onnx: {onnx_path}")
    if reload_onnx:
        onnx_model = onnx.load(onnx_path)
        print(f"Successfully LOADED onnx model to change input/output names: {onnx_path} ")
        
        if 'input_names' in params:
            print(f"Start to change input-names for {onnx_path} ")
            assert len(onnx_model.graph.input) == len(params['input_names']), RuntimeError(f"The number of onnx model input ({len(onnx_model.graph.input)}) must be same to input-names from yaml({len(params['input_names'])})")
            for idx, (input_tensor, input_name) in enumerate(zip(onnx_model.graph.input, params['input_names'])):
                print(f"  * {idx} changing {input_tensor.name} to {input_name}")
                input_tensor.name = input_name

        if 'output_names' in params:
            print(f"Start to change output names for {onnx_path} ")
            assert len(onnx_model.graph.output) == len(params['output_names']), RuntimeError(f"The number of onnx model output ({len(onnx_model.graph.output)}) must be same to output-names from yaml({len(params['output_names'])})")
            for jdx, (output_tensor, output_name) in enumerate(zip(onnx_model.graph.output, params['output_names'])):
                print(f"  * {jdx} changing {output_tensor.name} to {output_name}")
                output_tensor.name = output_name

            onnx.save(onnx_model, weights.replace(".pt", '_.onnx'))
            print(f"Successfully SAVED onnx model: {onnx_path} after changing input/output names")

if __name__ == '__main__':
    weights = "/DeepLearning/_projects/kt_g/wonchul/240807/train4/weights/best.pt"
    target = 'onnx'
    params = '/HDD/datasets/projects/mlearning/yolo/obb_detection/split_dataset_yolo_obb/export.yaml'
    
    export(weights, target, params)


# -*- coding: utf-8 -*-

from .checkpoint import load_checkpoint
from .layer_decay_optimizer_constructor import LayerDecayOptimizerConstructor
from .layer_decay_optimizer_constructor_backbone_frozen import LayerDecayOptimizerConstructorBackboneFronzen

from .checkpoint_hivit_1 import load_checkpoint_hivit_1
from .layer_decay_optimizer_constructor_hivit import HiViTLayerDecayOptimizerConstructor

__all__ = ['load_checkpoint', 'LayerDecayOptimizerConstructor', 'LayerDecayOptimizerConstructorBackboneFronzen']



import torch
from visionsuite.engines.classification.utils.registry import DATALOADERS
from visionsuite.engines.classification.utils.augment import get_mixup_cutmix
from torch.utils.data.dataloader import default_collate


@DATALOADERS.register()
def torch_dataloader(dataset, sampler, batch_size, workers, mixup_cutmix=None):
    
    collate_fn = default_collate
    if mixup_cutmix:
        mixup_cutmix = get_mixup_cutmix(
                mixup_alpha=mixup_cutmix['mixup_alpha'], 
                cutmix_alpha=mixup_cutmix['cutmix_alpha'], 
                num_classes=len(dataset.classes), 
                use_v2=mixup_cutmix['use_v2']
            )
        if mixup_cutmix is not None:
            def collate_fn(batch):
                return mixup_cutmix(*default_collate(batch))
            
        
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        num_workers=workers,
        pin_memory=True,
        collate_fn=collate_fn,
    )
    
    return dataloader




# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmcv.runner import BaseModule, auto_fp16
# from ..utils import ConvModule_Norm

# from ..builder import NECKS
from mmrotate.models.builder import ROTATED_NECKS


class ConvModule_Norm(ConvModule):
    def __init__(self, in_channels,
                out_channels,
                kernel, **kwargs):
        super().__init__(in_channels, out_channels, kernel, **kwargs)

        self.normType = kwargs.get('norm_cfg', {'type':''})
        if self.normType is not None:
            self.normType = self.normType['type']
    
    def forward(self, x, activate=True, norm=True):
        for layer in self.order:
            if layer == 'conv':
                if self.with_explicit_padding:
                    x = self.padding_layer(x)
                x = self.conv(x)
            elif layer == 'norm' and norm and self.with_norm:
                if 'LN' in self.normType:
                    x = x.permute(0, 2, 3, 1)
                    x = self.norm(x)
                    x = x.permute(0, 3, 1, 2).contiguous()
                else:
                    x = self.norm(x)
            elif layer == 'act' and activate and self.with_activation:
                x = self.activate(x)
        return x


@ROTATED_NECKS.register_module()
class SimpleFPN(BaseModule):
    r"""Feature Pyramid Network.
    This is an implementation of paper `Feature Pyramid Networks for Object
    Detection <https://arxiv.org/abs/1612.03144>`_.
    Args:
        in_channels (List[int]): Number of input channels per scale.
        out_channels (int): Number of output channels (used at each scale)
        num_outs (int): Number of output scales.
        start_level (int): Index of the start input backbone level used to
            build the feature pyramid. Default: 0.
        end_level (int): Index of the end input backbone level (exclusive) to
            build the feature pyramid. Default: -1, which means the last level.
        add_extra_convs (bool | str): If bool, it decides whether to add conv
            layers on top of the original feature maps. Default to False.
            If True, it is equivalent to `add_extra_convs='on_input'`.
            If str, it specifies the source feature map of the extra convs.
            Only the following options are allowed
            - 'on_input': Last feat map of neck inputs (i.e. backbone feature).
            - 'on_lateral':  Last feature map after lateral convs.
            - 'on_output': The last output feature map after fpn convs.
        relu_before_extra_convs (bool): Whether to apply relu before the extra
            conv. Default: False.
        no_norm_on_lateral (bool): Whether to apply norm on lateral.
            Default: False.
        conv_cfg (dict): Config dict for convolution layer. Default: None.
        norm_cfg (dict): Config dict for normalization layer. Default: None.
        act_cfg (str): Config dict for activation layer in ConvModule.
            Default: None.
        upsample_cfg (dict): Config dict for interpolate layer.
            Default: `dict(mode='nearest')`
        init_cfg (dict or list[dict], optional): Initialization config dict.
    Example:
        >>> import torch
        >>> in_channels = [2, 3, 5, 7]
        >>> scales = [340, 170, 84, 43]
        >>> inputs = [torch.rand(1, c, s, s)
        ...           for c, s in zip(in_channels, scales)]
        >>> self = FPN(in_channels, 11, len(in_channels)).eval()
        >>> outputs = self.forward(inputs)
        >>> for i in range(len(outputs)):
        ...     print(f'outputs[{i}].shape = {outputs[i].shape}')
        outputs[0].shape = torch.Size([1, 11, 340, 340])
        outputs[1].shape = torch.Size([1, 11, 170, 170])
        outputs[2].shape = torch.Size([1, 11, 84, 84])
        outputs[3].shape = torch.Size([1, 11, 43, 43])
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 num_outs,
                 start_level=0,
                 end_level=-1,
                 add_extra_convs=False,
                 relu_before_extra_convs=False,
                 no_norm_on_lateral=False,
                 conv_cfg=None,
                 norm_cfg=None,
                 act_cfg=None,
                 use_residual=True,
                 upsample_cfg=dict(mode='nearest'),
                 init_cfg=dict(
                     type='Xavier', layer='Conv2d', distribution='uniform')):
        super(SimpleFPN, self).__init__(init_cfg)
        assert isinstance(in_channels, list)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_ins = len(in_channels)
        self.num_outs = num_outs
        self.relu_before_extra_convs = relu_before_extra_convs
        self.no_norm_on_lateral = no_norm_on_lateral
        self.fp16_enabled = False
        self.upsample_cfg = upsample_cfg.copy()
        self.use_residual = use_residual

        if end_level == -1:
            self.backbone_end_level = self.num_ins
            assert num_outs >= self.num_ins - start_level
        else:
            # if end_level < inputs, no extra level is allowed
            self.backbone_end_level = end_level
            assert end_level <= len(in_channels)
            assert num_outs == end_level - start_level
        self.start_level = start_level
        self.end_level = end_level
        self.add_extra_convs = add_extra_convs
        assert isinstance(add_extra_convs, (str, bool))
        if isinstance(add_extra_convs, str):
            # Extra_convs_source choices: 'on_input', 'on_lateral', 'on_output'
            assert add_extra_convs in ('on_input', 'on_lateral', 'on_output')
        elif add_extra_convs:  # True
            self.add_extra_convs = 'on_input'

        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()

        for i in range(self.start_level, self.backbone_end_level):
            l_conv = ConvModule_Norm(
                in_channels[i],
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg if not self.no_norm_on_lateral else None,
                act_cfg=act_cfg,
                inplace=False)
            fpn_conv = ConvModule_Norm(
                out_channels,
                out_channels,
                3,
                padding=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
                inplace=False)

            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        # add extra conv layers (e.g., RetinaNet)
        extra_levels = num_outs - self.backbone_end_level + self.start_level
        if self.add_extra_convs and extra_levels >= 1:
            for i in range(extra_levels):
                if i == 0 and self.add_extra_convs == 'on_input':
                    in_channels = self.in_channels[self.backbone_end_level - 1]
                else:
                    in_channels = out_channels
                extra_fpn_conv = ConvModule_Norm(
                    in_channels,
                    out_channels,
                    3,
                    stride=2,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                    inplace=False)
                self.fpn_convs.append(extra_fpn_conv)

    @auto_fp16()
    def forward(self, inputs):
        """Forward function."""
        assert len(inputs) == len(self.in_channels)

        # build laterals
        laterals = [
            lateral_conv(inputs[i + self.start_level])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        # build top-down path
        used_backbone_levels = len(laterals)
        if self.use_residual:
            for i in range(used_backbone_levels - 1, 0, -1):
                # In some cases, fixing `scale factor` (e.g. 2) is preferred, but
                #  it cannot co-exist with `size` in `F.interpolate`.
                if 'scale_factor' in self.upsample_cfg:
                    laterals[i - 1] += F.interpolate(laterals[i],
                                                    **self.upsample_cfg)
                else:
                    prev_shape = laterals[i - 1].shape[2:]
                    laterals[i - 1] += F.interpolate(
                        laterals[i], size=prev_shape, **self.upsample_cfg)

        # build outputs
        # part 1: from original levels
        outs = [
            self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels)
        ]
        # part 2: add extra levels
        if self.num_outs > len(outs):
            # use max pool to get more levels on top of outputs
            # (e.g., Faster R-CNN, Mask R-CNN)
            if not self.add_extra_convs:
                for i in range(self.num_outs - used_backbone_levels):
                    outs.append(F.max_pool2d(outs[-1], 1, stride=2))
            # add conv layers on top of original feature maps (RetinaNet)
            else:
                if self.add_extra_convs == 'on_input':
                    extra_source = inputs[self.backbone_end_level - 1]
                elif self.add_extra_convs == 'on_lateral':
                    extra_source = laterals[-1]
                elif self.add_extra_convs == 'on_output':
                    extra_source = outs[-1]
                else:
                    raise NotImplementedError
                outs.append(self.fpn_convs[used_backbone_levels](extra_source))
                for i in range(used_backbone_levels + 1, self.num_outs):
                    if self.relu_before_extra_convs:
                        outs.append(self.fpn_convs[i](F.relu(outs[-1])))
                    else:
                        outs.append(self.fpn_convs[i](outs[-1]))
        return tuple(outs)


#!/usr/bin/env python3

import torch.nn as nn


def init_loss(loss, builder, num_classes=None):
    """Helper function to initialize loss (multilabel or single)."""
    if isinstance(loss, dict):
        if ("num_classes" in loss.keys()) and num_classes:
            loss.update(dict(num_classes=num_classes))
        out_loss = builder(loss)
    elif isinstance(loss, (list, tuple)):
        out_loss = nn.ModuleList()
        for l in loss:
            if ("num_classes" in l.keys()) and num_classes:
                l.update(dict(num_classes=num_classes))
            out_loss.append(builder(l))
    else:
        if loss is None:
            out_loss = None
        else:
            raise TypeError(
                f"loss must be a dict or sequence of dict or None,\
                but got {type(loss)}"
            )
    return out_loss


def get_loss_names(losses):
    loss_names = []

    if not isinstance(losses, (list, tuple)):
        losses = [losses]

    for loss in losses:
        if loss is not None:
            if not isinstance(loss, nn.ModuleList):
                loss_names.append(loss.loss_name)
            else:
                for l in loss:
                    loss_names.append(l.loss_name)

    return loss_names



import cv2
import json
import numpy as np
import os.path as osp

def get_filename(file_path, include_ext=False):
    if not include_ext:
        return osp.split(osp.splitext(file_path)[0])[-1]
    else:
        return osp.split(file_path)[-1]

def string_to_list_of_type(data, type_, lower=False, sep=','):

    if lower:
        return list(type_(value.strip().lower()) for value in data.split(sep)) if sep != "" else list(
            type_(value.strip().lower()) for value in list(data))
    else:
        return list(type_(value.strip())
                    for value in data.split(sep)) if sep != "" else list(type_(value.strip()) for value in list(data))

def letterbox(im, new_shape, color=(0, 0, 0)):
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border

    return im, ratio, (dw, dh)

class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(JsonEncoder, self).default(obj)
        
        
# def check_corrupt_image():
    


import os.path as osp
import psutil
import time 

from visionsuite.engines.utils.torch_utils.utils import save_on_master

def on_trainer_build_start(trainer, *args, **kwargs):
    pass        
     
def on_trainer_build_end(trainer, *args, **kwargs):
    
    for attribute_name in trainer.required_attributes:
        assert hasattr(trainer, attribute_name), ValueError(f'{attribute_name} must be assgined in trainer class')
        assert getattr(trainer, attribute_name) is not None, ValueError(f"{attribute_name} is None for trainer")

def on_trainer_epoch_start(trainer, *args, **kwargs):
    pass 

def on_trainer_epoch_end(trainer, *args, **kwargs):
    def _save_model():
        if trainer.archive and trainer.archive.weights_dir:
            checkpoint = {
                "model": trainer.model.model_without_ddp.state_dict(),
                "optimizer": trainer.optimizer.state_dict(),
                "lr_scheduler": trainer.lr_scheduler.state_dict(),
                "epoch": kwargs['epoch'],
                "args": trainer.args,
            }
            if trainer.model.model_ema:
                checkpoint["model_ema"] = trainer.model.model.model_ema.state_dict()
            if trainer.scaler:
                checkpoint["scaler"] = trainer.scaler.state_dict()
            
            if trainer.archive.args['model']['save_last']:
                save_on_master(checkpoint, osp.join(trainer.archive.weights_dir, "last.pth"))    
                
            if kwargs['epoch']%trainer.archive.args['model']['save_freq_epoch'] == 0:
                save_on_master(checkpoint, osp.join(trainer.archive.weights_dir, f"model_{kwargs['epoch']}.pth"))
    
    def _save_monitor():
        if trainer.archive and trainer.archive.monitor:
            trainer.archive.monitor.log({"learning rate": trainer.metric_logger.meters['lr'].value})
            trainer.archive.monitor.log({"train avg loss": trainer.metric_logger.meters['loss'].global_avg})
            trainer.archive.monitor.save()
  
    def _save_results():
        trainer.results.epoch = int(kwargs['epoch'])
        trainer.results.loss = float(round(trainer.metric_logger.meters['loss'].global_avg, 4))
        trainer.results.accuracy = float(round(trainer.metric_logger.meters["acc1"].global_avg, 4))
        trainer.results.learning_rate = float(round(trainer.metric_logger.meters['lr'].value, 4))
        trainer.results.cpu_usage = float(round(psutil.virtual_memory().used / 1024 / 1024 / 1024, 4))
        trainer.results.gpu_usage = trainer.gpu_logger.mean()
        trainer.results.time_for_a_epoch = float(round(time.time() - kwargs['start_time_epoch'], 3))
          
    _save_model()
    _save_monitor()
    _save_results()
    
    
def on_trainer_batch_start(trainer, *args, **kwargs):
    pass

def on_trainer_batch_end(trainer, *args, **kwargs):
    pass

def on_trainer_step_start(trainer, *args, **kwargs): # iteration for a batch
    pass

def on_trainer_step_end(trainer, *args, **kwargs): # iteration for a batch
    pass

callbacks = {
    "on_trainer_build_start": [on_trainer_build_start], 
    "on_trainer_build_end": [on_trainer_build_end], 
    "on_trainer_epoch_start": [on_trainer_epoch_start],
    "on_trainer_epoch_end": [on_trainer_epoch_end],
    "on_trainer_batch_start": [on_trainer_batch_start],
    "on_trainer_batch_end": [on_trainer_batch_end],
    "on_trainer_step_start": [on_trainer_step_start],
    "on_trainer_step_end": [on_trainer_step_end],
}




import random

import numpy as np
import torch
from torchvision import transforms as T
from torchvision.transforms import functional as F


def pad_if_smaller(img, size, fill=0):
    min_size = min(img.size)
    if min_size < size:
        ow, oh = img.size
        padh = size - oh if oh < size else 0
        padw = size - ow if ow < size else 0
        img = F.pad(img, (0, 0, padw, padh), fill=fill)
    return img


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target

class RandomResize:
    def __init__(self, min_size, max_size=None):
        self.min_size = min_size
        if max_size is None:
            max_size = min_size
        self.max_size = max_size

    def __call__(self, image, target):
        size = random.randint(self.min_size, self.max_size)
        image = F.resize(image, size, antialias=True)
        target = F.resize(target, size, interpolation=T.InterpolationMode.NEAREST)
        return image, target
    
class Resize:
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = F.resize(image, self.size, antialias=True)
        target = F.resize(target, self.size, interpolation=T.InterpolationMode.NEAREST)
        
        return image, target

class RandomHorizontalFlip:
    def __init__(self, flip_prob):
        self.flip_prob = flip_prob

    def __call__(self, image, target):
        if random.random() < self.flip_prob:
            image = F.hflip(image)
            target = F.hflip(target)
        return image, target


class RandomCrop:
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = pad_if_smaller(image, self.size)
        target = pad_if_smaller(target, self.size, fill=255)
        crop_params = T.RandomCrop.get_params(image, (self.size, self.size))
        image = F.crop(image, *crop_params)
        target = F.crop(target, *crop_params)
        return image, target


class CenterCrop:
    def __init__(self, size):
        self.size = size

    def __call__(self, image, target):
        image = F.center_crop(image, self.size)
        target = F.center_crop(target, self.size)
        return image, target


class PILToTensor:
    def __call__(self, image, target):
        image = F.pil_to_tensor(image)
        target = torch.as_tensor(np.array(target), dtype=torch.int64)
        return image, target


class ToDtype:
    def __init__(self, dtype, scale=False):
        self.dtype = dtype
        self.scale = scale

    def __call__(self, image, target):
        if not self.scale:
            return image.to(dtype=self.dtype), target
        image = F.convert_image_dtype(image, self.dtype)
        return image, target


class Normalize:
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, image, target):
        image = F.normalize(image, mean=self.mean, std=self.std)
        return image, target



#!/usr/bin/env python3

"""Deep Supervision Head

Supports binary and multilabel edges for supervision.
"""

import warnings
from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn
from mmcv.runner import BaseModule, auto_fp16, force_fp32
from mmseg.core import build_pixel_sampler
from mmseg.ops import resize

from ..builder import build_loss
from ..losses import calc_metrics
from .utils import init_loss


class BaseMultiSupervisionHead(BaseModule, metaclass=ABCMeta):
    def __init__(
        self,
        in_channels,
        channels,
        *,
        num_classes,
        edge_key,
        log_edge_keys,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
        in_index=-1,
        input_transform=None,
        binary_keys=[],
        multilabel_keys=[],
        binary_loss_weights=[],
        multilabel_loss_weights=[],
        loss_binary=None,
        loss_multilabel=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        ignore_index=255,
        sampler=None,
        align_corners=False,
        init_cfg=dict(type="Normal", std=0.01),
        no_accuracy=False,  # TODO: this is a hack to avoid memory errors
    ):
        super().__init__(init_cfg)

        assert isinstance(edge_key, str)
        self.edge_key = edge_key
        if log_edge_keys is None:
            log_edge_keys = []
        elif isinstance(log_edge_keys, str):
            log_edge_keys = [log_edge_keys]
        assert isinstance(log_edge_keys, (tuple, list))
        self.log_edge_keys = log_edge_keys

        self.loss_multilabel = init_loss(
            loss_multilabel, build_loss, num_classes=num_classes
        )
        if self.loss_multilabel is None:
            if len(multilabel_keys) > 0:
                warnings.warn(
                    "multilabel loss is None, but there seems to be some keys, "
                    f"{multilabel_keys},"
                    "removing keys..."
                )
                multilabel_keys = []
        else:
            if len(multilabel_loss_weights) == 0:
                multilabel_loss_weights = [1.0 for _ in multilabel_keys]

        self.loss_binary = init_loss(loss_binary, build_loss)
        if self.loss_binary is None:
            assert (
                self.loss_multilabel is not None
            ), "needs either binary or multilabel loss"
            if len(binary_keys) > 0:
                warnings.warn(
                    "binary loss is None, but there seems to be some keys, "
                    f"{binary_keys},"
                    "removing keys..."
                )
                binary_keys = []
        else:
            if len(binary_loss_weights) == 0:
                binary_loss_weights = [1.0 for _ in binary_keys]

        assert len(multilabel_keys) == len(multilabel_loss_weights)
        assert len(binary_keys) == len(binary_loss_weights)
        assert (
            len(multilabel_keys) + len(binary_keys) > 0
        ), f"there seems to be no edges to supervise: {multilabel_keys}, {binary_keys}"

        # loss info (key: weight)
        self.multilabel_info = dict(zip(multilabel_keys, multilabel_loss_weights))
        self.binary_info = dict(zip(binary_keys, binary_loss_weights))

        self._init_inputs(in_channels, in_index, input_transform)
        self.channels = channels
        self.num_classes = num_classes
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.in_index = in_index

        self.ignore_index = ignore_index
        self.align_corners = align_corners

        if sampler is not None:
            self.sampler = build_pixel_sampler(sampler, context=self)
        else:
            self.sampler = None

        self.fp16_enabled = False

        self.no_accuracy = no_accuracy

    def extra_repr(self):
        """Extra repr."""
        s = (
            f"input_transform={self.input_transform}, "
            f"ignore_index={self.ignore_index}, "
            f"align_corners={self.align_corners}"
        )
        return s

    def _init_inputs(self, in_channels, in_index, input_transform):
        """Check and initialize input transforms.

        The in_channels, in_index and input_transform must match.
        Specifically, when input_transform is None, only single feature map
        will be selected. So in_channels and in_index must be of type int.
        When input_transform

        Args:
            in_channels (int|Sequence[int]): Input channels.
            in_index (int|Sequence[int]): Input feature index.
            input_transform (str|None): Transformation type of input features.
                Options: 'resize_concat', 'multiple_select', None.
                'resize_concat': Multiple feature maps will be resize to the
                    same size as first one and than concat together.
                    Usually used in FCN head of HRNet.
                'multiple_select': Multiple feature maps will be bundle into
                    a list and passed into decode head.
                None: Only one select feature map is allowed.
        """

        if input_transform is not None:
            assert input_transform in ["resize_concat", "multiple_select"]
        self.input_transform = input_transform
        self.in_index = in_index
        if input_transform is not None:
            assert isinstance(in_channels, (list, tuple))
            assert isinstance(in_index, (list, tuple))
            assert len(in_channels) == len(in_index)
            if input_transform == "resize_concat":
                self.in_channels = sum(in_channels)
            else:
                self.in_channels = in_channels
        else:
            assert isinstance(in_channels, int)
            assert isinstance(in_index, int)
            self.in_channels = in_channels

    def _transform_inputs(self, inputs):
        """Transform inputs for decoder.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            Tensor: The transformed inputs
        """

        if self.input_transform == "resize_concat":
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(
                    input=x,
                    size=inputs[0].shape[2:],
                    mode="bilinear",
                    align_corners=self.align_corners,
                )
                for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == "multiple_select":
            inputs = [inputs[i] for i in self.in_index]
        else:
            inputs = inputs[self.in_index]

        return inputs

    @auto_fp16()
    @abstractmethod
    def forward(self, inputs, **kwargs):
        """Placeholder of forward function."""
        pass

    def forward_train(
        self,
        inputs,
        img_metas,
        gt_semantic_edge,
        train_cfg,
    ):
        logits = self(inputs)

        losses = dict()
        if self.multilabel_losses is not None:
            losses.update(self.multilabel_losses(logits, gt_semantic_edge))
        if self.binary_losses is not None:
            losses.update(self.binary_losses(logits, gt_semantic_edge))

        return losses

    def forward_test(self, inputs, img_metas, test_cfg):
        """use test_cfg's edge_key to choose which output"""
        edge_key = test_cfg.get("edge_key", self.edge_key)
        return self(inputs)[edge_key]

    @force_fp32(apply_to=("logits"))
    def binary_losses(self, logits, label):
        """Compute binary edge loss."""

        if isinstance(logits, torch.Tensor):
            # if a tensor is passed
            logits = dict(binary=logits)
        assert isinstance(logits, dict)

        loss = dict()

        # convert multilabel to binary edge, if needed
        if label.shape[1] != 1:
            # convert to binary
            label = (torch.sum(label, axis=1) > 0).unsqueeze(1).float()

        for k, logit in logits.items():
            if k in self.binary_info.keys():

                # get key specific weight
                w = self.binary_info[k]

                logit = resize(
                    input=logit,
                    size=label.shape[2:],  # (b, cls, h, w)
                    mode="bilinear",
                    align_corners=self.align_corners,
                )
                assert (
                    label.shape == logit.shape
                ), f"label, pred: {label.shape}, {logit.shape}"

                if not isinstance(self.loss_binary, nn.ModuleList):
                    losses_edge = [self.loss_binary]
                else:
                    losses_edge = self.loss_binary

                for loss_edge in losses_edge:
                    if loss_edge.loss_name not in loss:
                        loss[loss_edge.loss_name] = w * loss_edge(
                            logit,
                            label,
                            ignore_index=self.ignore_index,
                        )
                    else:
                        loss[loss_edge.loss_name] += w * loss_edge(
                            logit,
                            label,
                            ignore_index=self.ignore_index,
                        )

                if (k in self.log_edge_keys) and (not self.no_accuracy):
                    for name, v in calc_metrics(logit, label).items():
                        loss[k + "_" + name] = v

        return loss

    @force_fp32(apply_to=("logits"))
    def multilabel_losses(self, logits, label):
        """Compute multilabel edge loss."""

        if isinstance(logits, torch.Tensor):
            # if a tensor is passed
            logits = dict(multilabel=logits)
        assert isinstance(logits, dict)

        loss = dict()

        for k, logit in logits.items():
            if k in self.multilabel_info.keys():
                # get key specific weight
                w = self.multilabel_info[k]

                logit = resize(
                    input=logit,
                    size=label.shape[2:],  # (b, cls, h, w)
                    mode="bilinear",
                    align_corners=self.align_corners,
                )
                assert (
                    label.shape == logit.shape
                ), f"label, pred: {label.shape}, {logit.shape}"

                if not isinstance(self.loss_multilabel, nn.ModuleList):
                    losses_edge = [self.loss_multilabel]
                else:
                    losses_edge = self.loss_multilabel

                for loss_edge in losses_edge:
                    if loss_edge.loss_name not in loss:
                        loss[loss_edge.loss_name] = w * loss_edge(
                            logit,
                            label,
                            ignore_index=self.ignore_index,
                        )
                    else:
                        loss[loss_edge.loss_name] += w * loss_edge(
                            logit,
                            label,
                            ignore_index=self.ignore_index,
                        )

                if (k in self.log_edge_keys) and (not self.no_accuracy):
                    for name, v in calc_metrics(logit, label).items():
                        loss[k + "_" + name] = v

        return loss



import os
import math
import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint
from functools import partial
from collections import OrderedDict

from mmcv.runner import _load_checkpoint, load_state_dict
from mmdet.utils import get_root_logger
# from mmdet.models.builder import HEADS
from mmrotate.models.builder import ROTATED_HEADS
# from .bbox_head import BBoxHead
from mmrotate.models.roi_heads.bbox_heads.rotated_bbox_head import RotatedBBoxHead
from visionsuite.cores.spatial_transform_decoupling.models.vision_transformer import Block, trunc_normal_
# from ...utils.positional_encoding import get_2d_sincos_pos_embed


@ROTATED_HEADS.register_module()
class RotatedMAEBBoxHead(RotatedBBoxHead):
    def __init__(self,
                 in_channels,
                 img_size=224,
                 patch_size=16, 
                 embed_dim=256, 
                 depth=4,
                 num_heads=8, 
                 mlp_ratio=4., 
                 qkv_bias=True, 
                 qk_scale=None, 
                 drop_rate=0., 
                 attn_drop_rate=0.,
                 drop_path_rate=0.,
                 use_checkpoint=False,
                 init_cfg=None,
                 rois_mode=None,
                 **kwargs):
        super(RotatedMAEBBoxHead, self).__init__(init_cfg=init_cfg, **kwargs)

        if init_cfg is None: self.init_cfg = None
        self.rois_mode = rois_mode

        self.patch_size = patch_size
        self.use_checkpoint = use_checkpoint
        norm_layer = partial(nn.LayerNorm, eps=1e-6)
        num_patches = (img_size // patch_size) ** 2
        self.num_patches = num_patches

        # MAE decoder specifics
        self.norm = norm_layer(in_channels)
        self.decoder_embed = nn.Linear(in_channels, embed_dim, bias=True)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.decoder_blocks = nn.ModuleList([
            Block(
                embed_dim, num_heads, mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale, 
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth)
        ])
        self.decoder_pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim), requires_grad=False)  # fixed sin-cos embedding
        self.decoder_box_norm = norm_layer(embed_dim)
        # reconstruct fc_cls and fc_reg since input channels are changed
        if self.with_cls:
            self.fc_cls = nn.Linear(embed_dim, self.num_classes + 1)
        if self.with_reg:
            # out_dim_reg = 4 if self.reg_class_agnostic else 4 * self.num_classes
            out_dim_reg = 5 if self.reg_class_agnostic else 5 * self.num_classes # XYWHA
            self.fc_reg = nn.Linear(embed_dim, out_dim_reg)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        decoder_pos_embed = get_2d_sincos_pos_embed(self.decoder_pos_embed.shape[-1], int(self.num_patches**.5), cls_token=True)
        self.decoder_pos_embed.data.copy_(torch.from_numpy(decoder_pos_embed).float().unsqueeze(0))

    #def init_weights(self, pretrained):
    #    logger = get_root_logger()
    #    if os.path.isfile(pretrained):
    #        logger.info('loading checkpoint for {}'.format(self.__class__))
    #        checkpoint = _load_checkpoint(pretrained, map_location='cpu')
    #        if 'state_dict' in checkpoint:
    #            state_dict = checkpoint['state_dict']
    #        elif 'model' in checkpoint:
    #            state_dict = checkpoint['model']
    #        else:
    #            state_dict = checkpoint
    #        # TODO: match the decoder blocks, norm and head in the state_dict due to the different prefix
    #        new_state_dict = OrderedDict()
    #        for k, v in state_dict.items():
    #            if k.startswith('patch_embed') or k.startswith('blocks'):
    #                continue
    #            elif k in ['pos_embed']:
    #                continue
    #            else:
    #                new_state_dict[k] = v
    #        load_state_dict(self, new_state_dict, strict=False, logger=logger)
    #    else:
    #        raise ValueError(f"checkpoint path {pretrained} is invalid")

    def init_weights(self):
        logger = get_root_logger()
        if self.init_cfg is None:
            self.apply(self._init_weights)
        else:
            assert 'checkpoint' in self.init_cfg, f'Only support ' \
                                                  f'specify `Pretrained` in ' \
                                                  f'`init_cfg` in ' \
                                                  f'{self.__class__.__name__} '
            pretrained = self.init_cfg['checkpoint']
            if os.path.isfile(pretrained):
                logger.info('loading checkpoint for {}'.format(self.__class__))
                checkpoint = _load_checkpoint(pretrained, map_location='cpu')
                if 'state_dict' in checkpoint:
                    state_dict = checkpoint['state_dict']
                elif 'model' in checkpoint:
                    state_dict = checkpoint['model']
                else:
                    state_dict = checkpoint
                # TODO: match the decoder blocks, norm and head in the state_dict due to the different prefix
                new_state_dict = OrderedDict()
                for k, v in state_dict.items():
                    if k.startswith('patch_embed') or k.startswith('blocks'):
                        continue
                    elif k in ['pos_embed']:
                        continue
                    else:
                        new_state_dict[k] = v
                load_state_dict(self, new_state_dict, strict=False, logger=logger)
            else:
                raise ValueError(f"checkpoint path {pretrained} is invalid")

    def interpolate_pos_encoding(self, x, w, h):
        npatch = x.shape[1] - 1
        N = self.decoder_pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.decoder_pos_embed
        class_pos_embed = self.decoder_pos_embed[:, 0]
        patch_pos_embed = self.decoder_pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_size
        h0 = h // self.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
            mode='bicubic',
        )
        assert int(w0) == patch_pos_embed.shape[-2] and int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)
    
    def forward(self, x):
        B, C, W, H = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        x = self.decoder_embed(x)

        x = x + self.interpolate_pos_encoding(x, W * self.patch_size, H * self.patch_size)[:, 1:, :]
        for blk in self.decoder_blocks:
            if self.use_checkpoint:
                x = checkpoint(blk, x)
            else:
                x = blk(x)
        x = self.decoder_box_norm(x.mean(dim=1))
        cls_score = self.fc_cls(x) if self.with_cls else None
        bbox_pred = self.fc_reg(x) if self.with_reg else None
        return cls_score, bbox_pred


# --------------------------------------------------------
# 2D sine-cosine position embedding
# References:
# Transformer: https://github.com/tensorflow/models/blob/master/official/nlp/transformer/model_utils.py
# MoCo v3: https://github.com/facebookresearch/moco-v3
# --------------------------------------------------------
import numpy as np


def get_2d_sincos_pos_embed(embed_dim, grid_size, cls_token=False):
    """
    grid_size: int of the grid height and width
    return:
    pos_embed: [grid_size*grid_size, embed_dim] or [1+grid_size*grid_size, embed_dim] (w/ or w/o cls_token)
    """
    grid_h = np.arange(grid_size, dtype=np.float32)
    grid_w = np.arange(grid_size, dtype=np.float32)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, grid_size, grid_size])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed


def get_2d_sincos_pos_embed_from_grid(embed_dim, grid):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[0])  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[1])  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1) # (H*W, D)
    return emb


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=np.float)
    omega /= embed_dim / 2.
    omega = 1. / 10000**omega  # (D/2,)

    pos = pos.reshape(-1)  # (M,)
    out = np.einsum('m,d->md', pos, omega)  # (M, D/2), outer product

    emb_sin = np.sin(out) # (M, D/2)
    emb_cos = np.cos(out) # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb

